import time


class OAuth2Token(dict):
    def __init__(self, params):
        if params.get("expires_at"):
            try:
                params["expires_at"] = int(params["expires_at"])
            except ValueError:
                # If expires_at is not parseable, fall back to expires_in if available
                # Otherwise leave expires_at untouched
                if params.get("expires_in"):
                    params["expires_at"] = int(time.time()) + int(params["expires_in"])

        elif params.get("expires_in"):
            params["expires_at"] = int(time.time()) + int(params["expires_in"])

        super().__init__(params)

    def is_expired(self, leeway=60):
        expires_at = self.get("expires_at")
        if not expires_at:
            return None
        # Only check expiration if expires_at is an integer
        if not isinstance(expires_at, int):
            return None
        # small timedelta to consider token as expired before it actually expires
        expiration_threshold = expires_at - leeway
        return expiration_threshold < time.time()

    @classmethod
    def from_dict(cls, token):
        if isinstance(token, dict) and not isinstance(token, cls):
            token = cls(token)
        return token



import time
from unittest import mock

import pytest

from authlib.integrations.requests_client import AssertionSession


@pytest.fixture
def token():
    return {
        "token_type": "Bearer",
        "access_token": "a",
        "refresh_token": "b",
        "expires_in": "3600",
        "expires_at": int(time.time()) + 3600,
    }


def test_refresh_token(token):
    def verifier(r, **kwargs):
        resp = mock.MagicMock()
        resp.status_code = 200
        if r.url == "https://provider.test/token":
            assert "assertion=" in r.body
            resp.json = lambda: token
        return resp

    sess = AssertionSession(
        "https://provider.test/token",
        issuer="foo",
        subject="foo",
        audience="foo",
        alg="HS256",
        key="secret",
    )
    sess.send = verifier
    sess.get("https://provider.test")

    # trigger more case
    now = int(time.time())
    sess = AssertionSession(
        "https://provider.test/token",
        issuer="foo",
        subject=None,
        audience="foo",
        issued_at=now,
        expires_at=now + 3600,
        header={"alg": "HS256"},
        key="secret",
        scope="email",
        claims={"test_mode": "true"},
    )
    sess.send = verifier
    sess.get("https://provider.test")
    # trigger for branch test case
    sess.get("https://provider.test")


def test_without_alg():
    sess = AssertionSession(
        "https://provider.test/token",
        grant_type=AssertionSession.JWT_BEARER_GRANT_TYPE,
        issuer="foo",
        subject="foo",
        audience="foo",
        key="secret",
    )
    with pytest.raises(ValueError):
        sess.get("https://provider.test")



from ..rfc6749 import TokenMixin


class IntrospectionToken(dict, TokenMixin):
    def get_client_id(self):
        return self.get("client_id")

    def get_scope(self):
        return self.get("scope")

    def get_expires_in(self):
        # this method is only used in refresh token,
        # no need to implement it
        return 0

    def get_expires_at(self):
        return self.get("exp", 0)

    def __getattr__(self, key):
        # https://tools.ietf.org/html/rfc7662#section-2.2
        available_keys = {
            "active",
            "scope",
            "client_id",
            "username",
            "token_type",
            "exp",
            "iat",
            "nbf",
            "sub",
            "aud",
            "iss",
            "jti",
        }
        try:
            return object.__getattribute__(self, key)
        except AttributeError as error:
            if key in available_keys:
                return self.get(key)
            raise error



"""authlib.oauth2.rfc9068.token_validator.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Implementation of Validating JWT Access Tokens per `Section 4`_.

.. _`Section 7`: https://www.rfc-editor.org/rfc/rfc9068.html#name-validating-jwt-access-token
"""

from authlib.jose import jwt
from authlib.jose.errors import DecodeError
from authlib.jose.errors import JoseError
from authlib.oauth2.rfc6750.errors import InsufficientScopeError
from authlib.oauth2.rfc6750.errors import InvalidTokenError
from authlib.oauth2.rfc6750.validator import BearerTokenValidator

from .claims import JWTAccessTokenClaims


class JWTBearerTokenValidator(BearerTokenValidator):
    """JWTBearerTokenValidator can protect your resource server endpoints.

    :param issuer: The issuer from which tokens will be accepted.
    :param resource_server: An identifier for the current resource server,
        which must appear in the JWT ``aud`` claim.

    Developers needs to implement the missing methods::

        class MyJWTBearerTokenValidator(JWTBearerTokenValidator):
            def get_jwks(self): ...


        require_oauth = ResourceProtector()
        require_oauth.register_token_validator(
            MyJWTBearerTokenValidator(
                issuer="https://authorization-server.example.org",
                resource_server="https://resource-server.example.org",
            )
        )

    You can then protect resources depending on the JWT `scope`, `groups`,
    `roles` or `entitlements` claims::

        @require_oauth(
            scope="profile",
            groups="admins",
            roles="student",
            entitlements="captain",
        )
        def resource_endpoint(): ...
    """

    def __init__(self, issuer, resource_server, *args, **kwargs):
        self.issuer = issuer
        self.resource_server = resource_server
        super().__init__(*args, **kwargs)

    def get_jwks(self):
        """Return the JWKs that will be used to check the JWT access token signature.
        Developers MUST re-implement this method. Typically the JWKs are statically
        stored in the resource server configuration, or dynamically downloaded and
        cached using :ref:`specs/rfc8414`::

            def get_jwks(self):
                if "jwks" in cache:
                    return cache.get("jwks")

                server_metadata = get_server_metadata(self.issuer)
                jwks_uri = server_metadata.get("jwks_uri")
                cache["jwks"] = requests.get(jwks_uri).json()
                return cache["jwks"]
        """
        raise NotImplementedError()

    def validate_iss(self, claims, iss: "str") -> bool:
        # The issuer identifier for the authorization server (which is typically
        # obtained during discovery) MUST exactly match the value of the 'iss'
        # claim.
        return iss == self.issuer

    def authenticate_token(self, token_string):
        """"""
        # empty docstring avoids to display the irrelevant parent docstring

        claims_options = {
            "iss": {"essential": True, "validate": self.validate_iss},
            "exp": {"essential": True},
            "aud": {"essential": True, "value": self.resource_server},
            "sub": {"essential": True},
            "client_id": {"essential": True},
            "iat": {"essential": True},
            "jti": {"essential": True},
            "auth_time": {"essential": False},
            "acr": {"essential": False},
            "amr": {"essential": False},
            "scope": {"essential": False},
            "groups": {"essential": False},
            "roles": {"essential": False},
            "entitlements": {"essential": False},
        }
        jwks = self.get_jwks()

        # If the JWT access token is encrypted, decrypt it using the keys and algorithms
        # that the resource server specified during registration. If encryption was
        # negotiated with the authorization server at registration time and the incoming
        # JWT access token is not encrypted, the resource server SHOULD reject it.

        # The resource server MUST validate the signature of all incoming JWT access
        # tokens according to [RFC7515] using the algorithm specified in the JWT 'alg'
        # Header Parameter. The resource server MUST reject any JWT in which the value
        # of 'alg' is 'none'. The resource server MUST use the keys provided by the
        # authorization server.
        try:
            return jwt.decode(
                token_string,
                key=jwks,
                claims_cls=JWTAccessTokenClaims,
                claims_options=claims_options,
            )
        except DecodeError as exc:
            raise InvalidTokenError(
                realm=self.realm, extra_attributes=self.extra_attributes
            ) from exc

    def validate_token(
        self, token, scopes, request, groups=None, roles=None, entitlements=None
    ):
        """"""
        # empty docstring avoids to display the irrelevant parent docstring
        try:
            token.validate()
        except JoseError as exc:
            raise InvalidTokenError(
                realm=self.realm, extra_attributes=self.extra_attributes
            ) from exc

        # If an authorization request includes a scope parameter, the corresponding
        # issued JWT access token SHOULD include a 'scope' claim as defined in Section
        # 4.2 of [RFC8693]. All the individual scope strings in the 'scope' claim MUST
        # have meaning for the resources indicated in the 'aud' claim. See Section 5 for
        # more considerations about the relationship between scope strings and resources
        # indicated by the 'aud' claim.

        if self.scope_insufficient(token.get("scope", []), scopes):
            raise InsufficientScopeError()

        # Many authorization servers embed authorization attributes that go beyond the
        # delegated scenarios described by [RFC7519] in the access tokens they issue.
        # Typical examples include resource owner memberships in roles and groups that
        # are relevant to the resource being accessed, entitlements assigned to the
        # resource owner for the targeted resource that the authorization server knows
        # about, and so on. An authorization server wanting to include such attributes
        # in a JWT access token SHOULD use the 'groups', 'roles', and 'entitlements'
        # attributes of the 'User' resource schema defined by Section 4.1.2 of
        # [RFC7643]) as claim types.

        if self.scope_insufficient(token.get("groups"), groups):
            raise InvalidTokenError()

        if self.scope_insufficient(token.get("roles"), roles):
            raise InvalidTokenError()

        if self.scope_insufficient(token.get("entitlements"), entitlements):
            raise InvalidTokenError()



import json
import os

import pytest
from cryptography.exceptions import InvalidTag
from cryptography.hazmat.primitives.keywrap import InvalidUnwrap

from authlib.common.encoding import json_b64encode
from authlib.common.encoding import to_bytes
from authlib.common.encoding import to_unicode
from authlib.common.encoding import urlsafe_b64encode
from authlib.jose import JsonWebEncryption
from authlib.jose import OctKey
from authlib.jose import OKPKey
from authlib.jose import errors
from authlib.jose.drafts import register_jwe_draft
from authlib.jose.errors import DecodeError
from authlib.jose.errors import InvalidAlgorithmForMultipleRecipientsMode
from authlib.jose.errors import InvalidHeaderParameterNameError
from authlib.jose.util import extract_header
from tests.util import read_file_path

register_jwe_draft(JsonWebEncryption)


def test_not_enough_segments():
    s = "a.b.c"
    jwe = JsonWebEncryption()
    with pytest.raises(errors.DecodeError):
        jwe.deserialize_compact(s, None)


def test_invalid_header():
    jwe = JsonWebEncryption()
    public_key = read_file_path("rsa_public.pem")
    with pytest.raises(errors.MissingAlgorithmError):
        jwe.serialize_compact({}, "a", public_key)
    with pytest.raises(errors.UnsupportedAlgorithmError):
        jwe.serialize_compact(
            {"alg": "invalid"},
            "a",
            public_key,
        )
    with pytest.raises(errors.MissingEncryptionAlgorithmError):
        jwe.serialize_compact(
            {"alg": "RSA-OAEP"},
            "a",
            public_key,
        )
    with pytest.raises(errors.UnsupportedEncryptionAlgorithmError):
        jwe.serialize_compact(
            {"alg": "RSA-OAEP", "enc": "invalid"},
            "a",
            public_key,
        )
    with pytest.raises(errors.UnsupportedCompressionAlgorithmError):
        jwe.serialize_compact(
            {"alg": "RSA-OAEP", "enc": "A256GCM", "zip": "invalid"},
            "a",
            public_key,
        )


def test_not_supported_alg():
    public_key = read_file_path("rsa_public.pem")
    private_key = read_file_path("rsa_private.pem")

    jwe = JsonWebEncryption()
    s = jwe.serialize_compact(
        {"alg": "RSA-OAEP", "enc": "A256GCM"}, "hello", public_key
    )

    jwe = JsonWebEncryption(algorithms=["RSA1_5", "A256GCM"])
    with pytest.raises(errors.UnsupportedAlgorithmError):
        jwe.serialize_compact(
            {"alg": "RSA-OAEP", "enc": "A256GCM"},
            "hello",
            public_key,
        )
    with pytest.raises(errors.UnsupportedCompressionAlgorithmError):
        jwe.serialize_compact(
            {"alg": "RSA1_5", "enc": "A256GCM", "zip": "DEF"},
            "hello",
            public_key,
        )
    with pytest.raises(errors.UnsupportedAlgorithmError):
        jwe.deserialize_compact(
            s,
            private_key,
        )

    jwe = JsonWebEncryption(algorithms=["RSA-OAEP", "A192GCM"])
    with pytest.raises(errors.UnsupportedEncryptionAlgorithmError):
        jwe.serialize_compact(
            {"alg": "RSA-OAEP", "enc": "A256GCM"},
            "hello",
            public_key,
        )
    with pytest.raises(errors.UnsupportedCompressionAlgorithmError):
        jwe.serialize_compact(
            {"alg": "RSA-OAEP", "enc": "A192GCM", "zip": "DEF"},
            "hello",
            public_key,
        )
    with pytest.raises(errors.UnsupportedEncryptionAlgorithmError):
        jwe.deserialize_compact(
            s,
            private_key,
        )


def test_inappropriate_sender_key_for_serialize_compact():
    jwe = JsonWebEncryption()
    alice_key = {
        "kty": "EC",
        "crv": "P-256",
        "x": "WKn-ZIGevcwGIyyrzFoZNBdaq9_TsqzGl96oc0CWuis",
        "y": "y77t-RvAHRKTsSGdIYUfweuOvwrvDD-Q3Hv5J0fSKbE",
        "d": "Hndv7ZZjs_ke8o9zXYo3iq-Yr8SewI5vrqd0pAvEPqg",
    }
    bob_key = {
        "kty": "EC",
        "crv": "P-256",
        "x": "weNJy2HscCSM6AEDTDg04biOvhFhyyWvOHQfeF_PxMQ",
        "y": "e8lnCO-AlStT-NJVX-crhB7QRYhiix03illJOVAOyck",
        "d": "VEmDZpDXXK8p8N0Cndsxs924q6nS1RXFASRl6BfUqdw",
    }

    protected = {"alg": "ECDH-1PU", "enc": "A256GCM"}
    with pytest.raises(ValueError):
        jwe.serialize_compact(protected, b"hello", bob_key)

    protected = {"alg": "ECDH-ES", "enc": "A256GCM"}
    with pytest.raises(ValueError):
        jwe.serialize_compact(
            protected,
            b"hello",
            bob_key,
            sender_key=alice_key,
        )


def test_inappropriate_sender_key_for_deserialize_compact():
    jwe = JsonWebEncryption()
    alice_key = {
        "kty": "EC",
        "crv": "P-256",
        "x": "WKn-ZIGevcwGIyyrzFoZNBdaq9_TsqzGl96oc0CWuis",
        "y": "y77t-RvAHRKTsSGdIYUfweuOvwrvDD-Q3Hv5J0fSKbE",
        "d": "Hndv7ZZjs_ke8o9zXYo3iq-Yr8SewI5vrqd0pAvEPqg",
    }
    bob_key = {
        "kty": "EC",
        "crv": "P-256",
        "x": "weNJy2HscCSM6AEDTDg04biOvhFhyyWvOHQfeF_PxMQ",
        "y": "e8lnCO-AlStT-NJVX-crhB7QRYhiix03illJOVAOyck",
        "d": "VEmDZpDXXK8p8N0Cndsxs924q6nS1RXFASRl6BfUqdw",
    }

    protected = {"alg": "ECDH-1PU", "enc": "A256GCM"}
    data = jwe.serialize_compact(protected, b"hello", bob_key, sender_key=alice_key)
    with pytest.raises(ValueError):
        jwe.deserialize_compact(data, bob_key)

    protected = {"alg": "ECDH-ES", "enc": "A256GCM"}
    data = jwe.serialize_compact(protected, b"hello", bob_key)
    with pytest.raises(ValueError):
        jwe.deserialize_compact(data, bob_key, sender_key=alice_key)


def test_compact_rsa():
    jwe = JsonWebEncryption()
    s = jwe.serialize_compact(
        {"alg": "RSA-OAEP", "enc": "A256GCM"},
        "hello",
        read_file_path("rsa_public.pem"),
    )
    data = jwe.deserialize_compact(s, read_file_path("rsa_private.pem"))
    header, payload = data["header"], data["payload"]
    assert payload == b"hello"
    assert header["alg"] == "RSA-OAEP"


def test_with_zip_header():
    jwe = JsonWebEncryption()
    s = jwe.serialize_compact(
        {"alg": "RSA-OAEP", "enc": "A128CBC-HS256", "zip": "DEF"},
        "hello",
        read_file_path("rsa_public.pem"),
    )
    data = jwe.deserialize_compact(s, read_file_path("rsa_private.pem"))
    header, payload = data["header"], data["payload"]
    assert payload == b"hello"
    assert header["alg"] == "RSA-OAEP"


def test_aes_jwe():
    jwe = JsonWebEncryption()
    sizes = [128, 192, 256]
    _enc_choices = [
        "A128CBC-HS256",
        "A192CBC-HS384",
        "A256CBC-HS512",
        "A128GCM",
        "A192GCM",
        "A256GCM",
    ]
    for s in sizes:
        alg = f"A{s}KW"
        key = os.urandom(s // 8)
        for enc in _enc_choices:
            protected = {"alg": alg, "enc": enc}
            data = jwe.serialize_compact(protected, b"hello", key)
            rv = jwe.deserialize_compact(data, key)
            assert rv["payload"] == b"hello"


def test_aes_jwe_invalid_key():
    jwe = JsonWebEncryption()
    protected = {"alg": "A128KW", "enc": "A128GCM"}
    with pytest.raises(ValueError):
        jwe.serialize_compact(protected, b"hello", b"invalid-key")


def test_aes_gcm_jwe():
    jwe = JsonWebEncryption()
    sizes = [128, 192, 256]
    _enc_choices = [
        "A128CBC-HS256",
        "A192CBC-HS384",
        "A256CBC-HS512",
        "A128GCM",
        "A192GCM",
        "A256GCM",
    ]
    for s in sizes:
        alg = f"A{s}GCMKW"
        key = os.urandom(s // 8)
        for enc in _enc_choices:
            protected = {"alg": alg, "enc": enc}
            data = jwe.serialize_compact(protected, b"hello", key)
            rv = jwe.deserialize_compact(data, key)
            assert rv["payload"] == b"hello"


def test_aes_gcm_jwe_invalid_key():
    jwe = JsonWebEncryption()
    protected = {"alg": "A128GCMKW", "enc": "A128GCM"}
    with pytest.raises(ValueError):
        jwe.serialize_compact(protected, b"hello", b"invalid-key")


def test_serialize_compact_fails_if_header_contains_unknown_field_while_private_fields_restricted():
    jwe = JsonWebEncryption(private_headers=set())
    key = OKPKey.generate_key("X25519", is_private=True)

    protected = {"alg": "ECDH-ES+A128KW", "enc": "A128GCM", "foo": "bar"}

    with pytest.raises(InvalidHeaderParameterNameError):
        jwe.serialize_compact(
            protected,
            b"hello",
            key,
        )


def test_serialize_compact_allows_unknown_fields_in_header_while_private_fields_not_restricted():
    jwe = JsonWebEncryption()
    key = OKPKey.generate_key("X25519", is_private=True)

    protected = {"alg": "ECDH-ES+A128KW", "enc": "A128GCM", "foo": "bar"}

    data = jwe.serialize_compact(protected, b"hello", key)
    rv = jwe.deserialize_compact(data, key)
    assert rv["payload"] == b"hello"


def test_serialize_json_fails_if_protected_header_contains_unknown_field_while_private_fields_restricted():
    jwe = JsonWebEncryption(private_headers=set())
    key = OKPKey.generate_key("X25519", is_private=True)

    protected = {"alg": "ECDH-ES+A128KW", "enc": "A128GCM", "foo": "bar"}
    header_obj = {"protected": protected}

    with pytest.raises(InvalidHeaderParameterNameError):
        jwe.serialize_json(
            header_obj,
            b"hello",
            key,
        )


def test_serialize_json_fails_if_unprotected_header_contains_unknown_field_while_private_fields_restricted():
    jwe = JsonWebEncryption(private_headers=set())
    key = OKPKey.generate_key("X25519", is_private=True)

    protected = {"alg": "ECDH-ES+A128KW", "enc": "A128GCM"}
    unprotected = {"foo": "bar"}
    header_obj = {"protected": protected, "unprotected": unprotected}

    with pytest.raises(InvalidHeaderParameterNameError):
        jwe.serialize_json(
            header_obj,
            b"hello",
            key,
        )


def test_serialize_json_fails_if_recipient_header_contains_unknown_field_while_private_fields_restricted():
    jwe = JsonWebEncryption(private_headers=set())
    key = OKPKey.generate_key("X25519", is_private=True)

    protected = {"alg": "ECDH-ES+A128KW", "enc": "A128GCM"}
    recipients = [{"header": {"foo": "bar"}}]
    header_obj = {"protected": protected, "recipients": recipients}

    with pytest.raises(InvalidHeaderParameterNameError):
        jwe.serialize_json(
            header_obj,
            b"hello",
            key,
        )


def test_serialize_json_allows_unknown_fields_in_headers_while_private_fields_not_restricted():
    jwe = JsonWebEncryption()
    key = OKPKey.generate_key("X25519", is_private=True)

    protected = {"alg": "ECDH-ES+A128KW", "enc": "A128GCM", "foo1": "bar1"}
    unprotected = {"foo2": "bar2"}
    recipients = [{"header": {"foo3": "bar3"}}]
    header_obj = {
        "protected": protected,
        "unprotected": unprotected,
        "recipients": recipients,
    }

    data = jwe.serialize_json(header_obj, b"hello", key)
    rv = jwe.deserialize_json(data, key)
    assert rv["payload"] == b"hello"


def test_serialize_json_ignores_additional_members_in_recipients_elements():
    jwe = JsonWebEncryption()
    key = OKPKey.generate_key("X25519", is_private=True)

    protected = {"alg": "ECDH-ES+A128KW", "enc": "A128GCM"}

    data = jwe.serialize_compact(protected, b"hello", key)
    rv = jwe.deserialize_compact(data, key)
    assert rv["payload"] == b"hello"


def test_deserialize_json_fails_if_protected_header_contains_unknown_field_while_private_fields_restricted():
    jwe = JsonWebEncryption(private_headers=set())
    key = OKPKey.generate_key("X25519", is_private=True)

    protected = {"alg": "ECDH-ES+A128KW", "enc": "A128GCM"}
    header_obj = {"protected": protected}

    data = jwe.serialize_json(header_obj, b"hello", key)

    decoded_protected = extract_header(to_bytes(data["protected"]), DecodeError)
    decoded_protected["foo"] = "bar"
    data["protected"] = to_unicode(json_b64encode(decoded_protected))

    with pytest.raises(InvalidHeaderParameterNameError):
        jwe.deserialize_json(data, key)


def test_deserialize_json_fails_if_unprotected_header_contains_unknown_field_while_private_fields_restricted():
    jwe = JsonWebEncryption(private_headers=set())
    key = OKPKey.generate_key("X25519", is_private=True)

    protected = {"alg": "ECDH-ES+A128KW", "enc": "A128GCM"}
    header_obj = {"protected": protected}

    data = jwe.serialize_json(header_obj, b"hello", key)

    data["unprotected"] = {"foo": "bar"}

    with pytest.raises(InvalidHeaderParameterNameError):
        jwe.deserialize_json(data, key)


def test_deserialize_json_fails_if_recipient_header_contains_unknown_field_while_private_fields_restricted():
    jwe = JsonWebEncryption(private_headers=set())
    key = OKPKey.generate_key("X25519", is_private=True)

    protected = {"alg": "ECDH-ES+A128KW", "enc": "A128GCM"}
    header_obj = {"protected": protected}

    data = jwe.serialize_json(header_obj, b"hello", key)

    data["recipients"][0]["header"] = {"foo": "bar"}

    with pytest.raises(InvalidHeaderParameterNameError):
        jwe.deserialize_json(data, key)


def test_deserialize_json_allows_unknown_fields_in_headers_while_private_fields_not_restricted():
    jwe = JsonWebEncryption()
    key = OKPKey.generate_key("X25519", is_private=True)

    protected = {"alg": "ECDH-ES+A128KW", "enc": "A128GCM"}
    header_obj = {"protected": protected}

    data = jwe.serialize_json(header_obj, b"hello", key)

    data["unprotected"] = {"foo1": "bar1"}
    data["recipients"][0]["header"] = {"foo2": "bar2"}

    rv = jwe.deserialize_json(data, key)
    assert rv["payload"] == b"hello"


def test_deserialize_json_ignores_additional_members_in_recipients_elements():
    jwe = JsonWebEncryption()
    key = OKPKey.generate_key("X25519", is_private=True)

    protected = {"alg": "ECDH-ES+A128KW", "enc": "A128GCM"}
    header_obj = {"protected": protected}

    data = jwe.serialize_json(header_obj, b"hello", key)

    data["recipients"][0]["foo"] = "bar"

    data = jwe.serialize_compact(protected, b"hello", key)
    rv = jwe.deserialize_compact(data, key)
    assert rv["payload"] == b"hello"


def test_deserialize_json_ignores_additional_members_in_jwe_message():
    jwe = JsonWebEncryption()
    key = OKPKey.generate_key("X25519", is_private=True)

    protected = {"alg": "ECDH-ES+A128KW", "enc": "A128GCM"}
    header_obj = {"protected": protected}

    data = jwe.serialize_json(header_obj, b"hello", key)

    data["foo"] = "bar"

    data = jwe.serialize_compact(protected, b"hello", key)
    rv = jwe.deserialize_compact(data, key)
    assert rv["payload"] == b"hello"


def test_ecdh_es_key_agreement_computation():
    # https://tools.ietf.org/html/rfc7518#appendix-C
    alice_ephemeral_key = {
        "kty": "EC",
        "crv": "P-256",
        "x": "gI0GAILBdu7T53akrFmMyGcsF3n5dO7MmwNBHKW5SV0",
        "y": "SLW_xSffzlPWrHEVI30DHM_4egVwt3NQqeUD7nMFpps",
        "d": "0_NxaRPUMQoAJt50Gz8YiTr8gRTwyEaCumd-MToTmIo",
    }
    bob_static_key = {
        "kty": "EC",
        "crv": "P-256",
        "x": "weNJy2HscCSM6AEDTDg04biOvhFhyyWvOHQfeF_PxMQ",
        "y": "e8lnCO-AlStT-NJVX-crhB7QRYhiix03illJOVAOyck",
        "d": "VEmDZpDXXK8p8N0Cndsxs924q6nS1RXFASRl6BfUqdw",
    }

    headers = {
        "alg": "ECDH-ES",
        "enc": "A128GCM",
        "apu": "QWxpY2U",
        "apv": "Qm9i",
        "epk": {
            "kty": "EC",
            "crv": "P-256",
            "x": "gI0GAILBdu7T53akrFmMyGcsF3n5dO7MmwNBHKW5SV0",
            "y": "SLW_xSffzlPWrHEVI30DHM_4egVwt3NQqeUD7nMFpps",
        },
    }

    alg = JsonWebEncryption.ALG_REGISTRY["ECDH-ES"]
    enc = JsonWebEncryption.ENC_REGISTRY["A128GCM"]

    alice_ephemeral_key = alg.prepare_key(alice_ephemeral_key)
    bob_static_key = alg.prepare_key(bob_static_key)

    alice_ephemeral_pubkey = alice_ephemeral_key.get_op_key("wrapKey")
    bob_static_pubkey = bob_static_key.get_op_key("wrapKey")

    # Derived key computation at Alice

    # Step-by-step methods verification
    _shared_key_at_alice = alice_ephemeral_key.exchange_shared_key(bob_static_pubkey)
    assert _shared_key_at_alice == bytes(
        [
            158,
            86,
            217,
            29,
            129,
            113,
            53,
            211,
            114,
            131,
            66,
            131,
            191,
            132,
            38,
            156,
            251,
            49,
            110,
            163,
            218,
            128,
            106,
            72,
            246,
            218,
            167,
            121,
            140,
            254,
            144,
            196,
        ]
    )

    _fixed_info_at_alice = alg.compute_fixed_info(headers, enc.key_size)
    assert _fixed_info_at_alice == bytes(
        [
            0,
            0,
            0,
            7,
            65,
            49,
            50,
            56,
            71,
            67,
            77,
            0,
            0,
            0,
            5,
            65,
            108,
            105,
            99,
            101,
            0,
            0,
            0,
            3,
            66,
            111,
            98,
            0,
            0,
            0,
            128,
        ]
    )

    _dk_at_alice = alg.compute_derived_key(
        _shared_key_at_alice, _fixed_info_at_alice, enc.key_size
    )
    assert _dk_at_alice == bytes(
        [86, 170, 141, 234, 248, 35, 109, 32, 92, 34, 40, 205, 113, 167, 16, 26]
    )
    assert urlsafe_b64encode(_dk_at_alice) == b"VqqN6vgjbSBcIijNcacQGg"

    # All-in-one method verification
    dk_at_alice = alg.deliver(
        alice_ephemeral_key, bob_static_pubkey, headers, enc.key_size
    )
    assert dk_at_alice == bytes(
        [86, 170, 141, 234, 248, 35, 109, 32, 92, 34, 40, 205, 113, 167, 16, 26]
    )
    assert urlsafe_b64encode(dk_at_alice) == b"VqqN6vgjbSBcIijNcacQGg"

    # Derived key computation at Bob

    # Step-by-step methods verification
    _shared_key_at_bob = bob_static_key.exchange_shared_key(alice_ephemeral_pubkey)
    assert _shared_key_at_bob == _shared_key_at_alice

    _fixed_info_at_bob = alg.compute_fixed_info(headers, enc.key_size)
    assert _fixed_info_at_bob == _fixed_info_at_alice

    _dk_at_bob = alg.compute_derived_key(
        _shared_key_at_bob, _fixed_info_at_bob, enc.key_size
    )
    assert _dk_at_bob == _dk_at_alice

    # All-in-one method verification
    dk_at_bob = alg.deliver(
        bob_static_key, alice_ephemeral_pubkey, headers, enc.key_size
    )
    assert dk_at_bob == dk_at_alice


def test_ecdh_es_jwe_in_direct_key_agreement_mode():
    jwe = JsonWebEncryption()
    key = {
        "kty": "EC",
        "crv": "P-256",
        "x": "weNJy2HscCSM6AEDTDg04biOvhFhyyWvOHQfeF_PxMQ",
        "y": "e8lnCO-AlStT-NJVX-crhB7QRYhiix03illJOVAOyck",
        "d": "VEmDZpDXXK8p8N0Cndsxs924q6nS1RXFASRl6BfUqdw",
    }

    for enc in [
        "A128CBC-HS256",
        "A192CBC-HS384",
        "A256CBC-HS512",
        "A128GCM",
        "A192GCM",
        "A256GCM",
    ]:
        protected = {"alg": "ECDH-ES", "enc": enc}
        data = jwe.serialize_compact(protected, b"hello", key)
        rv = jwe.deserialize_compact(data, key)
        assert rv["payload"] == b"hello"


def test_ecdh_es_jwe_json_serialization_single_recipient_in_direct_key_agreement_mode():
    jwe = JsonWebEncryption()
    key = OKPKey.generate_key("X25519", is_private=True)

    protected = {"alg": "ECDH-ES", "enc": "A128GCM"}
    header_obj = {"protected": protected}
    data = jwe.serialize_json(header_obj, b"hello", key)
    rv = jwe.deserialize_json(data, key)
    assert rv["payload"] == b"hello"


def test_ecdh_es_jwe_in_key_agreement_with_key_wrapping_mode():
    jwe = JsonWebEncryption()
    key = {
        "kty": "EC",
        "crv": "P-256",
        "x": "weNJy2HscCSM6AEDTDg04biOvhFhyyWvOHQfeF_PxMQ",
        "y": "e8lnCO-AlStT-NJVX-crhB7QRYhiix03illJOVAOyck",
        "d": "VEmDZpDXXK8p8N0Cndsxs924q6nS1RXFASRl6BfUqdw",
    }

    for alg in [
        "ECDH-ES+A128KW",
        "ECDH-ES+A192KW",
        "ECDH-ES+A256KW",
    ]:
        for enc in [
            "A128CBC-HS256",
            "A192CBC-HS384",
            "A256CBC-HS512",
            "A128GCM",
            "A192GCM",
            "A256GCM",
        ]:
            protected = {"alg": alg, "enc": enc}
            data = jwe.serialize_compact(protected, b"hello", key)
            rv = jwe.deserialize_compact(data, key)
            assert rv["payload"] == b"hello"


def test_ecdh_es_jwe_with_okp_key_in_direct_key_agreement_mode():
    jwe = JsonWebEncryption()
    key = OKPKey.generate_key("X25519", is_private=True)

    for enc in [
        "A128CBC-HS256",
        "A192CBC-HS384",
        "A256CBC-HS512",
        "A128GCM",
        "A192GCM",
        "A256GCM",
    ]:
        protected = {"alg": "ECDH-ES", "enc": enc}
        data = jwe.serialize_compact(protected, b"hello", key)
        rv = jwe.deserialize_compact(data, key)
        assert rv["payload"] == b"hello"


def test_ecdh_es_jwe_with_okp_key_in_key_agreement_with_key_wrapping_mode():
    jwe = JsonWebEncryption()
    key = OKPKey.generate_key("X25519", is_private=True)

    for alg in [
        "ECDH-ES+A128KW",
        "ECDH-ES+A192KW",
        "ECDH-ES+A256KW",
    ]:
        for enc in [
            "A128CBC-HS256",
            "A192CBC-HS384",
            "A256CBC-HS512",
            "A128GCM",
            "A192GCM",
            "A256GCM",
        ]:
            protected = {"alg": alg, "enc": enc}
            data = jwe.serialize_compact(protected, b"hello", key)
            rv = jwe.deserialize_compact(data, key)
            assert rv["payload"] == b"hello"


def test_ecdh_es_jwe_with_json_serialization_when_kid_is_not_specified():
    jwe = JsonWebEncryption()

    bob_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X25519",
            "x": "BT7aR0ItXfeDAldeeOlXL_wXqp-j5FltT0vRSG16kRw",
            "d": "1gDirl_r_Y3-qUa3WXHgEXrrEHngWThU3c9zj9A2uBg",
        }
    )
    charlie_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X25519",
            "x": "q-LsvU772uV_2sPJhfAIq-3vnKNVefNoIlvyvg1hrnE",
            "d": "Jcv8gklhMjC0b-lsk5onBbppWAx5ncNtbM63Jr9xBQE",
        }
    )

    protected = {
        "alg": "ECDH-ES+A256KW",
        "enc": "A256GCM",
        "apu": "QWxpY2U",
        "apv": "Qm9iIGFuZCBDaGFybGll",
    }

    unprotected = {"jku": "https://provider.test/jwks"}

    recipients = [
        {"header": {"kid": "bob-key-2"}},
        {"header": {"kid": "2021-05-06"}},
    ]

    jwe_aad = b"Authenticate me too."

    header_obj = {
        "protected": protected,
        "unprotected": unprotected,
        "recipients": recipients,
        "aad": jwe_aad,
    }

    payload = b"Three is a magic number."

    data = jwe.serialize_json(header_obj, payload, [bob_key, charlie_key])

    rv_at_bob = jwe.deserialize_json(data, bob_key)

    assert rv_at_bob["header"]["protected"].keys() == protected.keys() | {"epk"}
    assert {
        k: rv_at_bob["header"]["protected"][k]
        for k in rv_at_bob["header"]["protected"].keys() - {"epk"}
    } == protected
    assert rv_at_bob["header"]["unprotected"] == unprotected
    assert rv_at_bob["header"]["recipients"] == recipients
    assert rv_at_bob["header"]["aad"] == jwe_aad
    assert rv_at_bob["payload"] == payload

    rv_at_charlie = jwe.deserialize_json(data, charlie_key)

    assert rv_at_charlie["header"]["protected"].keys() == protected.keys() | {"epk"}
    assert {
        k: rv_at_charlie["header"]["protected"][k]
        for k in rv_at_charlie["header"]["protected"].keys() - {"epk"}
    } == protected
    assert rv_at_charlie["header"]["unprotected"] == unprotected
    assert rv_at_charlie["header"]["recipients"] == recipients
    assert rv_at_charlie["header"]["aad"] == jwe_aad
    assert rv_at_charlie["payload"] == payload


def test_ecdh_es_jwe_with_json_serialization_when_kid_is_specified():
    jwe = JsonWebEncryption()

    bob_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X25519",
            "kid": "bob-key-2",
            "x": "BT7aR0ItXfeDAldeeOlXL_wXqp-j5FltT0vRSG16kRw",
            "d": "1gDirl_r_Y3-qUa3WXHgEXrrEHngWThU3c9zj9A2uBg",
        }
    )
    charlie_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X25519",
            "kid": "2021-05-06",
            "x": "q-LsvU772uV_2sPJhfAIq-3vnKNVefNoIlvyvg1hrnE",
            "d": "Jcv8gklhMjC0b-lsk5onBbppWAx5ncNtbM63Jr9xBQE",
        }
    )

    protected = {
        "alg": "ECDH-ES+A256KW",
        "enc": "A256GCM",
        "apu": "QWxpY2U",
        "apv": "Qm9iIGFuZCBDaGFybGll",
    }

    unprotected = {"jku": "https://provider.test/jwks"}

    recipients = [
        {"header": {"kid": "bob-key-2"}},
        {"header": {"kid": "2021-05-06"}},
    ]

    jwe_aad = b"Authenticate me too."

    header_obj = {
        "protected": protected,
        "unprotected": unprotected,
        "recipients": recipients,
        "aad": jwe_aad,
    }

    payload = b"Three is a magic number."

    data = jwe.serialize_json(header_obj, payload, [bob_key, charlie_key])

    rv_at_bob = jwe.deserialize_json(data, bob_key)

    assert rv_at_bob["header"]["protected"].keys() == protected.keys() | {"epk"}
    assert {
        k: rv_at_bob["header"]["protected"][k]
        for k in rv_at_bob["header"]["protected"].keys() - {"epk"}
    } == protected
    assert rv_at_bob["header"]["unprotected"] == unprotected
    assert rv_at_bob["header"]["recipients"] == recipients
    assert rv_at_bob["header"]["aad"] == jwe_aad
    assert rv_at_bob["payload"] == payload

    rv_at_charlie = jwe.deserialize_json(data, charlie_key)

    assert rv_at_charlie["header"]["protected"].keys() == protected.keys() | {"epk"}
    assert {
        k: rv_at_charlie["header"]["protected"][k]
        for k in rv_at_charlie["header"]["protected"].keys() - {"epk"}
    } == protected
    assert rv_at_charlie["header"]["unprotected"] == unprotected
    assert rv_at_charlie["header"]["recipients"] == recipients
    assert rv_at_charlie["header"]["aad"] == jwe_aad
    assert rv_at_charlie["payload"] == payload


def test_ecdh_es_jwe_with_json_serialization_for_single_recipient():
    jwe = JsonWebEncryption()

    key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X25519",
            "x": "BT7aR0ItXfeDAldeeOlXL_wXqp-j5FltT0vRSG16kRw",
            "d": "1gDirl_r_Y3-qUa3WXHgEXrrEHngWThU3c9zj9A2uBg",
        }
    )

    protected = {
        "alg": "ECDH-ES+A256KW",
        "enc": "A256GCM",
        "apu": "QWxpY2U",
        "apv": "Qm9i",
    }

    unprotected = {"jku": "https://provider.test/jwks"}

    recipients = [{"header": {"kid": "bob-key-2"}}]

    jwe_aad = b"Authenticate me too."

    header_obj = {
        "protected": protected,
        "unprotected": unprotected,
        "recipients": recipients,
        "aad": jwe_aad,
    }

    payload = b"Three is a magic number."

    data = jwe.serialize_json(header_obj, payload, key)

    rv = jwe.deserialize_json(data, key)

    assert rv["header"]["protected"].keys() == protected.keys() | {"epk"}
    assert {
        k: rv["header"]["protected"][k]
        for k in rv["header"]["protected"].keys() - {"epk"}
    } == protected
    assert rv["header"]["unprotected"] == unprotected
    assert rv["header"]["recipients"] == recipients
    assert rv["header"]["aad"] == jwe_aad
    assert rv["payload"] == payload


def test_ecdh_es_encryption_fails_json_serialization_multiple_recipients_in_direct_key_agreement_mode():
    jwe = JsonWebEncryption()
    bob_key = OKPKey.generate_key("X25519", is_private=True)
    charlie_key = OKPKey.generate_key("X25519", is_private=True)

    protected = {"alg": "ECDH-ES", "enc": "A128GCM"}
    header_obj = {"protected": protected}
    with pytest.raises(InvalidAlgorithmForMultipleRecipientsMode):
        jwe.serialize_json(
            header_obj,
            b"hello",
            [bob_key, charlie_key],
        )


def test_ecdh_es_decryption_with_public_key_fails():
    jwe = JsonWebEncryption()
    protected = {"alg": "ECDH-ES", "enc": "A128GCM"}

    key = {
        "kty": "EC",
        "crv": "P-256",
        "x": "weNJy2HscCSM6AEDTDg04biOvhFhyyWvOHQfeF_PxMQ",
        "y": "e8lnCO-AlStT-NJVX-crhB7QRYhiix03illJOVAOyck",
    }
    data = jwe.serialize_compact(protected, b"hello", key)
    with pytest.raises(ValueError):
        jwe.deserialize_compact(data, key)


def test_ecdh_es_encryption_fails_if_key_curve_is_inappropriate():
    jwe = JsonWebEncryption()
    protected = {"alg": "ECDH-ES", "enc": "A128GCM"}

    key = OKPKey.generate_key("Ed25519", is_private=False)
    with pytest.raises(ValueError):
        jwe.serialize_compact(protected, b"hello", key)


def test_ecdh_es_decryption_fails_if_key_matches_to_no_recipient():
    jwe = JsonWebEncryption()

    bob_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X25519",
            "x": "BT7aR0ItXfeDAldeeOlXL_wXqp-j5FltT0vRSG16kRw",
            "d": "1gDirl_r_Y3-qUa3WXHgEXrrEHngWThU3c9zj9A2uBg",
        }
    )
    charlie_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X25519",
            "x": "q-LsvU772uV_2sPJhfAIq-3vnKNVefNoIlvyvg1hrnE",
            "d": "Jcv8gklhMjC0b-lsk5onBbppWAx5ncNtbM63Jr9xBQE",
        }
    )

    protected = {
        "alg": "ECDH-ES+A256KW",
        "enc": "A256GCM",
        "apu": "QWxpY2U",
        "apv": "Qm9i",
    }

    unprotected = {"jku": "https://provider.test/jwks"}

    recipients = [{"header": {"kid": "bob-key-2"}}]

    jwe_aad = b"Authenticate me too."

    header_obj = {
        "protected": protected,
        "unprotected": unprotected,
        "recipients": recipients,
        "aad": jwe_aad,
    }

    payload = b"Three is a magic number."

    data = jwe.serialize_json(header_obj, payload, bob_key)

    with pytest.raises(InvalidUnwrap):
        jwe.deserialize_json(data, charlie_key)


def test_decryption_with_json_serialization_succeeds_while_encrypted_key_for_another_recipient_is_invalid():
    jwe = JsonWebEncryption()

    alice_key = OKPKey.import_key(
        {
            "kid": "Alice's key",
            "kty": "OKP",
            "crv": "X25519",
            "x": "Knbm_BcdQr7WIoz-uqit9M0wbcfEr6y-9UfIZ8QnBD4",
            "d": "i9KuFhSzEBsiv3PKVL5115OCdsqQai5nj_Flzfkw5jU",
        }
    )
    OKPKey.import_key(
        {
            "kid": "Bob's key",
            "kty": "OKP",
            "crv": "X25519",
            "x": "BT7aR0ItXfeDAldeeOlXL_wXqp-j5FltT0vRSG16kRw",
            "d": "1gDirl_r_Y3-qUa3WXHgEXrrEHngWThU3c9zj9A2uBg",
        }
    )
    charlie_key = OKPKey.import_key(
        {
            "kid": "Charlie's key",
            "kty": "OKP",
            "crv": "X25519",
            "x": "q-LsvU772uV_2sPJhfAIq-3vnKNVefNoIlvyvg1hrnE",
            "d": "Jcv8gklhMjC0b-lsk5onBbppWAx5ncNtbM63Jr9xBQE",
        }
    )

    data = {
        "protected": "eyJhbGciOiJFQ0RILTFQVStBMTI4S1ciLCJlbmMiOiJBMjU2Q0JDLUhTNTEyIiwiYXB1Ijoi"
        + "UVd4cFkyVSIsImFwdiI6IlFtOWlJR0Z1WkNCRGFHRnliR2xsIiwiZXBrIjp7Imt0eSI6Ik9L"
        + "UCIsImNydiI6IlgyNTUxOSIsIngiOiJrOW9mX2NwQWFqeTBwb1c1Z2FpeFhHczluSGt3ZzFB"
        + "RnFVQUZhMzlkeUJjIn19",
        "unprotected": {"jku": "https://provider.test/jwks"},
        "recipients": [
            {
                "header": {"kid": "Bob's key"},
                "encrypted_key": "pOMVA9_PtoRe7xXW1139NzzN1UhiFoio8lGto9cf0t8PyU-sjNXH8-LIRLycq8CHJQbDwvQ"
                + "eU1cSl55cQ0hGezJu2N9IY0QM",  # Invalid encrypted key
            },
            {
                "header": {"kid": "Charlie's key"},
                "encrypted_key": "56GVudgRLIMEElQ7DpXsijJVRSWUSDNdbWkdV3g0GUNq6hcT_GkxwnxlPIWrTXCqRpVKQC8"
                + "fe4z3PQ2YH2afvjQ28aiCTWFE",  # Valid encrypted key
            },
        ],
        "iv": "AAECAwQFBgcICQoLDA0ODw",
        "ciphertext": "Az2IWsISEMDJvyc5XRL-3-d-RgNBOGolCsxFFoUXFYw",
        "tag": "HLb4fTlm8spGmij3RyOs2gJ4DpHM4hhVRwdF_hGb3WQ",
    }

    rv_at_charlie = jwe.deserialize_json(data, charlie_key, sender_key=alice_key)

    assert rv_at_charlie.keys() == {"header", "payload"}

    assert rv_at_charlie["header"].keys() == {
        "protected",
        "unprotected",
        "recipients",
    }

    assert rv_at_charlie["header"]["protected"] == {
        "alg": "ECDH-1PU+A128KW",
        "enc": "A256CBC-HS512",
        "apu": "QWxpY2U",
        "apv": "Qm9iIGFuZCBDaGFybGll",
        "epk": {
            "kty": "OKP",
            "crv": "X25519",
            "x": "k9of_cpAajy0poW5gaixXGs9nHkwg1AFqUAFa39dyBc",
        },
    }

    assert rv_at_charlie["header"]["unprotected"] == {
        "jku": "https://provider.test/jwks"
    }

    assert rv_at_charlie["header"]["recipients"] == [
        {"header": {"kid": "Bob's key"}},
        {"header": {"kid": "Charlie's key"}},
    ]

    assert rv_at_charlie["payload"] == b"Three is a magic number."


def test_decryption_with_json_serialization_fails_if_encrypted_key_for_this_recipient_is_invalid():
    jwe = JsonWebEncryption()

    alice_key = OKPKey.import_key(
        {
            "kid": "Alice's key",
            "kty": "OKP",
            "crv": "X25519",
            "x": "Knbm_BcdQr7WIoz-uqit9M0wbcfEr6y-9UfIZ8QnBD4",
            "d": "i9KuFhSzEBsiv3PKVL5115OCdsqQai5nj_Flzfkw5jU",
        }
    )
    bob_key = OKPKey.import_key(
        {
            "kid": "Bob's key",
            "kty": "OKP",
            "crv": "X25519",
            "x": "BT7aR0ItXfeDAldeeOlXL_wXqp-j5FltT0vRSG16kRw",
            "d": "1gDirl_r_Y3-qUa3WXHgEXrrEHngWThU3c9zj9A2uBg",
        }
    )
    OKPKey.import_key(
        {
            "kid": "Charlie's key",
            "kty": "OKP",
            "crv": "X25519",
            "x": "q-LsvU772uV_2sPJhfAIq-3vnKNVefNoIlvyvg1hrnE",
            "d": "Jcv8gklhMjC0b-lsk5onBbppWAx5ncNtbM63Jr9xBQE",
        }
    )

    data = {
        "protected": "eyJhbGciOiJFQ0RILTFQVStBMTI4S1ciLCJlbmMiOiJBMjU2Q0JDLUhTNTEyIiwiYXB1Ijoi"
        + "UVd4cFkyVSIsImFwdiI6IlFtOWlJR0Z1WkNCRGFHRnliR2xsIiwiZXBrIjp7Imt0eSI6Ik9L"
        + "UCIsImNydiI6IlgyNTUxOSIsIngiOiJrOW9mX2NwQWFqeTBwb1c1Z2FpeFhHczluSGt3ZzFB"
        + "RnFVQUZhMzlkeUJjIn19",
        "unprotected": {"jku": "https://provider.test/jwks"},
        "recipients": [
            {
                "header": {"kid": "Bob's key"},
                "encrypted_key": "pOMVA9_PtoRe7xXW1139NzzN1UhiFoio8lGto9cf0t8PyU-sjNXH8-LIRLycq8CHJQbDwvQ"
                + "eU1cSl55cQ0hGezJu2N9IY0QM",  # Invalid encrypted key
            },
            {
                "header": {"kid": "Charlie's key"},
                "encrypted_key": "56GVudgRLIMEElQ7DpXsijJVRSWUSDNdbWkdV3g0GUNq6hcT_GkxwnxlPIWrTXCqRpVKQC8"
                + "fe4z3PQ2YH2afvjQ28aiCTWFE",  # Valid encrypted key
            },
        ],
        "iv": "AAECAwQFBgcICQoLDA0ODw",
        "ciphertext": "Az2IWsISEMDJvyc5XRL-3-d-RgNBOGolCsxFFoUXFYw",
        "tag": "HLb4fTlm8spGmij3RyOs2gJ4DpHM4hhVRwdF_hGb3WQ",
    }

    with pytest.raises(InvalidUnwrap):
        jwe.deserialize_json(data, bob_key, sender_key=alice_key)


def test_dir_alg():
    jwe = JsonWebEncryption()
    key = OctKey.generate_key(128, is_private=True)
    protected = {"alg": "dir", "enc": "A128GCM"}
    data = jwe.serialize_compact(protected, b"hello", key)
    rv = jwe.deserialize_compact(data, key)
    assert rv["payload"] == b"hello"

    key2 = OctKey.generate_key(256, is_private=True)
    with pytest.raises(ValueError):
        jwe.deserialize_compact(data, key2)

    with pytest.raises(ValueError):
        jwe.serialize_compact(protected, b"hello", key2)


def test_decryption_of_message_to_multiple_recipients_by_matching_key():
    jwe = JsonWebEncryption()

    alice_public_key = OKPKey.import_key(
        {
            "kid": "WjKgJV7VRw3hmgU6--4v15c0Aewbcvat1BsRFTIqa5Q",
            "kty": "OKP",
            "crv": "X25519",
            "x": "Knbm_BcdQr7WIoz-uqit9M0wbcfEr6y-9UfIZ8QnBD4",
        }
    )

    key_store = {}

    charlie_X448_key_id = "did:example:123#_TKzHv2jFIyvdTGF1Dsgwngfdg3SH6TpDv0Ta1aOEkw"
    charlie_X448_key = OKPKey.import_key(
        {
            "kid": "_TKzHv2jFIyvdTGF1Dsgwngfdg3SH6TpDv0Ta1aOEkw",
            "kty": "OKP",
            "crv": "X448",
            "x": "M-OMugy74ksznVQ-Bp6MC_-GEPSrT8yiAtminJvw0j_UxJtpNHl_hcWMSf_Pfm_ws0vVWvAfwwA",
            "d": "VGZPkclj_7WbRaRMzBqxpzXIpc2xz1d3N1ay36UxdVLfKaP33hABBMpddTRv1f-hRsQUNvmlGOg",
        }
    )
    key_store[charlie_X448_key_id] = charlie_X448_key

    charlie_X25519_key_id = (
        "did:example:123#ZC2jXTO6t4R501bfCXv3RxarZyUbdP2w_psLwMuY6ec"
    )
    charlie_X25519_key = OKPKey.import_key(
        {
            "kid": "ZC2jXTO6t4R501bfCXv3RxarZyUbdP2w_psLwMuY6ec",
            "kty": "OKP",
            "crv": "X25519",
            "x": "q-LsvU772uV_2sPJhfAIq-3vnKNVefNoIlvyvg1hrnE",
            "d": "Jcv8gklhMjC0b-lsk5onBbppWAx5ncNtbM63Jr9xBQE",
        }
    )
    key_store[charlie_X25519_key_id] = charlie_X25519_key

    data = """
        {
            "protected": "eyJhbGciOiJFQ0RILTFQVStBMTI4S1ciLCJlbmMiOiJBMjU2Q0JDLUhTNTEyIiwiYXB1IjoiUVd4cFkyVSIsImFwdiI6IlFtOWlJR0Z1WkNCRGFHRnliR2xsIiwiZXBrIjp7Imt0eSI6Ik9LUCIsImNydiI6IlgyNTUxOSIsIngiOiJrOW9mX2NwQWFqeTBwb1c1Z2FpeFhHczluSGt3ZzFBRnFVQUZhMzlkeUJjIn19",
            "unprotected": {
                "jku": "https://provider.test/jwks"
            },
            "recipients": [
                {
                    "header": {
                        "kid": "did:example:123#_Qq0UL2Fq651Q0Fjd6TvnYE-faHiOpRlPVQcY_-tA4A"
                    },
                    "encrypted_key": "pOMVA9_PtoRe7xXW1139NzzN1UhiFoio8lGto9cf0t8PyU-sjNXH8-LIRLycq8CHJQbDwvQeU1cSl55cQ0hGezJu2N9IY0QN"
                },
                {
                    "header": {
                        "kid": "did:example:123#ZC2jXTO6t4R501bfCXv3RxarZyUbdP2w_psLwMuY6ec"
                    },
                    "encrypted_key": "56GVudgRLIMEElQ7DpXsijJVRSWUSDNdbWkdV3g0GUNq6hcT_GkxwnxlPIWrTXCqRpVKQC8fe4z3PQ2YH2afvjQ28aiCTWFE"
                }
            ],
            "iv": "AAECAwQFBgcICQoLDA0ODw",
            "ciphertext": "Az2IWsISEMDJvyc5XRL-3-d-RgNBOGolCsxFFoUXFYw",
            "tag": "HLb4fTlm8spGmij3RyOs2gJ4DpHM4hhVRwdF_hGb3WQ"
        }"""

    parsed_data = jwe.parse_json(data)

    available_key_id = next(
        recipient["header"]["kid"]
        for recipient in parsed_data["recipients"]
        if recipient["header"]["kid"] in key_store.keys()
    )
    available_key = key_store[available_key_id]

    rv = jwe.deserialize_json(
        parsed_data, (available_key_id, available_key), sender_key=alice_public_key
    )

    assert rv.keys() == {"header", "payload"}

    assert rv["header"].keys() == {"protected", "unprotected", "recipients"}

    assert rv["header"]["protected"] == {
        "alg": "ECDH-1PU+A128KW",
        "enc": "A256CBC-HS512",
        "apu": "QWxpY2U",
        "apv": "Qm9iIGFuZCBDaGFybGll",
        "epk": {
            "kty": "OKP",
            "crv": "X25519",
            "x": "k9of_cpAajy0poW5gaixXGs9nHkwg1AFqUAFa39dyBc",
        },
    }

    assert rv["header"]["unprotected"] == {"jku": "https://provider.test/jwks"}

    assert rv["header"]["recipients"] == [
        {
            "header": {
                "kid": "did:example:123#_Qq0UL2Fq651Q0Fjd6TvnYE-faHiOpRlPVQcY_-tA4A"
            }
        },
        {
            "header": {
                "kid": "did:example:123#ZC2jXTO6t4R501bfCXv3RxarZyUbdP2w_psLwMuY6ec"
            }
        },
    ]

    assert rv["payload"] == b"Three is a magic number."


def test_decryption_of_json_string():
    jwe = JsonWebEncryption()

    alice_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X25519",
            "x": "Knbm_BcdQr7WIoz-uqit9M0wbcfEr6y-9UfIZ8QnBD4",
            "d": "i9KuFhSzEBsiv3PKVL5115OCdsqQai5nj_Flzfkw5jU",
        }
    )
    bob_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X25519",
            "x": "BT7aR0ItXfeDAldeeOlXL_wXqp-j5FltT0vRSG16kRw",
            "d": "1gDirl_r_Y3-qUa3WXHgEXrrEHngWThU3c9zj9A2uBg",
        }
    )
    charlie_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X25519",
            "x": "q-LsvU772uV_2sPJhfAIq-3vnKNVefNoIlvyvg1hrnE",
            "d": "Jcv8gklhMjC0b-lsk5onBbppWAx5ncNtbM63Jr9xBQE",
        }
    )

    data = """
        {
            "protected": "eyJhbGciOiJFQ0RILTFQVStBMTI4S1ciLCJlbmMiOiJBMjU2Q0JDLUhTNTEyIiwiYXB1IjoiUVd4cFkyVSIsImFwdiI6IlFtOWlJR0Z1WkNCRGFHRnliR2xsIiwiZXBrIjp7Imt0eSI6Ik9LUCIsImNydiI6IlgyNTUxOSIsIngiOiJrOW9mX2NwQWFqeTBwb1c1Z2FpeFhHczluSGt3ZzFBRnFVQUZhMzlkeUJjIn19",
            "unprotected": {
                "jku": "https://provider.test/jwks"
            },
            "recipients": [
                {
                    "header": {
                        "kid": "bob-key-2"
                    },
                    "encrypted_key": "pOMVA9_PtoRe7xXW1139NzzN1UhiFoio8lGto9cf0t8PyU-sjNXH8-LIRLycq8CHJQbDwvQeU1cSl55cQ0hGezJu2N9IY0QN"
                },
                {
                    "header": {
                        "kid": "2021-05-06"
                    },
                    "encrypted_key": "56GVudgRLIMEElQ7DpXsijJVRSWUSDNdbWkdV3g0GUNq6hcT_GkxwnxlPIWrTXCqRpVKQC8fe4z3PQ2YH2afvjQ28aiCTWFE"
                }
            ],
            "iv": "AAECAwQFBgcICQoLDA0ODw",
            "ciphertext": "Az2IWsISEMDJvyc5XRL-3-d-RgNBOGolCsxFFoUXFYw",
            "tag": "HLb4fTlm8spGmij3RyOs2gJ4DpHM4hhVRwdF_hGb3WQ"
        }"""

    rv_at_bob = jwe.deserialize_json(data, bob_key, sender_key=alice_key)

    assert rv_at_bob.keys() == {"header", "payload"}

    assert rv_at_bob["header"].keys() == {"protected", "unprotected", "recipients"}

    assert rv_at_bob["header"]["protected"] == {
        "alg": "ECDH-1PU+A128KW",
        "enc": "A256CBC-HS512",
        "apu": "QWxpY2U",
        "apv": "Qm9iIGFuZCBDaGFybGll",
        "epk": {
            "kty": "OKP",
            "crv": "X25519",
            "x": "k9of_cpAajy0poW5gaixXGs9nHkwg1AFqUAFa39dyBc",
        },
    }

    assert rv_at_bob["header"]["unprotected"] == {"jku": "https://provider.test/jwks"}

    assert rv_at_bob["header"]["recipients"] == [
        {"header": {"kid": "bob-key-2"}},
        {"header": {"kid": "2021-05-06"}},
    ]

    assert rv_at_bob["payload"] == b"Three is a magic number."

    rv_at_charlie = jwe.deserialize_json(data, charlie_key, sender_key=alice_key)

    assert rv_at_charlie.keys() == {"header", "payload"}

    assert rv_at_charlie["header"].keys() == {
        "protected",
        "unprotected",
        "recipients",
    }

    assert rv_at_charlie["header"]["protected"] == {
        "alg": "ECDH-1PU+A128KW",
        "enc": "A256CBC-HS512",
        "apu": "QWxpY2U",
        "apv": "Qm9iIGFuZCBDaGFybGll",
        "epk": {
            "kty": "OKP",
            "crv": "X25519",
            "x": "k9of_cpAajy0poW5gaixXGs9nHkwg1AFqUAFa39dyBc",
        },
    }

    assert rv_at_charlie["header"]["unprotected"] == {
        "jku": "https://provider.test/jwks"
    }

    assert rv_at_charlie["header"]["recipients"] == [
        {"header": {"kid": "bob-key-2"}},
        {"header": {"kid": "2021-05-06"}},
    ]

    assert rv_at_charlie["payload"] == b"Three is a magic number."


def test_parse_json():
    json_msg = """
        {
            "protected": "eyJhbGciOiJFQ0RILTFQVStBMTI4S1ciLCJlbmMiOiJBMjU2Q0JDLUhTNTEyIiwiYXB1IjoiUVd4cFkyVSIsImFwdiI6IlFtOWlJR0Z1WkNCRGFHRnliR2xsIiwiZXBrIjp7Imt0eSI6Ik9LUCIsImNydiI6IlgyNTUxOSIsIngiOiJrOW9mX2NwQWFqeTBwb1c1Z2FpeFhHczluSGt3ZzFBRnFVQUZhMzlkeUJjIn19",
            "unprotected": {
                "jku": "https://provider.test/jwks"
            },
            "recipients": [
                {
                    "header": {
                        "kid": "bob-key-2"
                    },
                    "encrypted_key": "pOMVA9_PtoRe7xXW1139NzzN1UhiFoio8lGto9cf0t8PyU-sjNXH8-LIRLycq8CHJQbDwvQeU1cSl55cQ0hGezJu2N9IY0QN"
                },
                {
                    "header": {
                        "kid": "2021-05-06"
                    },
                    "encrypted_key": "56GVudgRLIMEElQ7DpXsijJVRSWUSDNdbWkdV3g0GUNq6hcT_GkxwnxlPIWrTXCqRpVKQC8fe4z3PQ2YH2afvjQ28aiCTWFE"
                }
            ],
            "iv": "AAECAwQFBgcICQoLDA0ODw",
            "ciphertext": "Az2IWsISEMDJvyc5XRL-3-d-RgNBOGolCsxFFoUXFYw",
            "tag": "HLb4fTlm8spGmij3RyOs2gJ4DpHM4hhVRwdF_hGb3WQ"
        }"""

    parsed_msg = JsonWebEncryption.parse_json(json_msg)

    assert parsed_msg == {
        "protected": "eyJhbGciOiJFQ0RILTFQVStBMTI4S1ciLCJlbmMiOiJBMjU2Q0JDLUhTNTEyIiwiYXB1IjoiUVd4cFkyVSIsImFwdiI6IlFtOWlJR0Z1WkNCRGFHRnliR2xsIiwiZXBrIjp7Imt0eSI6Ik9LUCIsImNydiI6IlgyNTUxOSIsIngiOiJrOW9mX2NwQWFqeTBwb1c1Z2FpeFhHczluSGt3ZzFBRnFVQUZhMzlkeUJjIn19",
        "unprotected": {"jku": "https://provider.test/jwks"},
        "recipients": [
            {
                "header": {"kid": "bob-key-2"},
                "encrypted_key": "pOMVA9_PtoRe7xXW1139NzzN1UhiFoio8lGto9cf0t8PyU-sjNXH8-LIRLycq8CHJQbDwvQeU1cSl55cQ0hGezJu2N9IY0QN",
            },
            {
                "header": {"kid": "2021-05-06"},
                "encrypted_key": "56GVudgRLIMEElQ7DpXsijJVRSWUSDNdbWkdV3g0GUNq6hcT_GkxwnxlPIWrTXCqRpVKQC8fe4z3PQ2YH2afvjQ28aiCTWFE",
            },
        ],
        "iv": "AAECAwQFBgcICQoLDA0ODw",
        "ciphertext": "Az2IWsISEMDJvyc5XRL-3-d-RgNBOGolCsxFFoUXFYw",
        "tag": "HLb4fTlm8spGmij3RyOs2gJ4DpHM4hhVRwdF_hGb3WQ",
    }


def test_parse_json_fails_if_json_msg_is_invalid():
    json_msg = """
        {
            "protected": "eyJhbGciOiJFQ0RILTFQVStBMTI4S1ciLCJlbmMiOiJBMjU2Q0JDLUhTNTEyIiwiYXB1IjoiUVd4cFkyVSIsImFwdiI6IlFtOWlJR0Z1WkNCRGFHRnliR2xsIiwiZXBrIjp7Imt0eSI6Ik9LUCIsImNydiI6IlgyNTUxOSIsIngiOiJrOW9mX2NwQWFqeTBwb1c1Z2FpeFhHczluSGt3ZzFBRnFVQUZhMzlkeUJjIn19",
            "unprotected": {
                "jku": "https://provider.test/jwks"
            },
            "recipients": [
                {
                    "header": {
                        "kid": "bob-key-2"
                    ,
                    "encrypted_key": "pOMVA9_PtoRe7xXW1139NzzN1UhiFoio8lGto9cf0t8PyU-sjNXH8-LIRLycq8CHJQbDwvQeU1cSl55cQ0hGezJu2N9IY0QN"
                },
                {
                    "header": {
                        "kid": "2021-05-06"
                    },
                    "encrypted_key": "56GVudgRLIMEElQ7DpXsijJVRSWUSDNdbWkdV3g0GUNq6hcT_GkxwnxlPIWrTXCqRpVKQC8fe4z3PQ2YH2afvjQ28aiCTWFE"
                }
            ],
            "iv": "AAECAwQFBgcICQoLDA0ODw",
            "ciphertext": "Az2IWsISEMDJvyc5XRL-3-d-RgNBOGolCsxFFoUXFYw",
            "tag": "HLb4fTlm8spGmij3RyOs2gJ4DpHM4hhVRwdF_hGb3WQ"
        }"""

    with pytest.raises(DecodeError):
        JsonWebEncryption.parse_json(json_msg)


def test_decryption_fails_if_ciphertext_is_invalid():
    jwe = JsonWebEncryption()

    alice_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X25519",
            "x": "Knbm_BcdQr7WIoz-uqit9M0wbcfEr6y-9UfIZ8QnBD4",
            "d": "i9KuFhSzEBsiv3PKVL5115OCdsqQai5nj_Flzfkw5jU",
        }
    )
    bob_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X25519",
            "x": "BT7aR0ItXfeDAldeeOlXL_wXqp-j5FltT0vRSG16kRw",
            "d": "1gDirl_r_Y3-qUa3WXHgEXrrEHngWThU3c9zj9A2uBg",
        }
    )

    data = {
        "protected": "eyJhbGciOiJFQ0RILTFQVStBMTI4S1ciLCJlbmMiOiJBMjU2Q0JDLUhTNTEyIiwiYXB1Ijoi"
        + "UVd4cFkyVSIsImFwdiI6IlFtOWlJR0Z1WkNCRGFHRnliR2xsIiwiZXBrIjp7Imt0eSI6Ik9L"
        + "UCIsImNydiI6IlgyNTUxOSIsIngiOiJrOW9mX2NwQWFqeTBwb1c1Z2FpeFhHczluSGt3ZzFB"
        + "RnFVQUZhMzlkeUJjIn19",
        "unprotected": {"jku": "https://provider.test/jwks"},
        "recipients": [
            {
                "header": {"kid": "bob-key-2"},
                "encrypted_key": "pOMVA9_PtoRe7xXW1139NzzN1UhiFoio8lGto9cf0t8PyU-sjNXH8-LIRLycq8CHJQbDwvQ"
                + "eU1cSl55cQ0hGezJu2N9IY0QN",
            }
        ],
        "iv": "AAECAwQFBgcICQoLDA0ODw",
        "ciphertext": "Az2IWsISEMDJvyc5XRL-3-d-RgNBOGolCsxFFoUXFY",  # invalid ciphertext
        "tag": "HLb4fTlm8spGmij3RyOs2gJ4DpHM4hhVRwdF_hGb3WQ",
    }

    with pytest.raises(InvalidTag):
        jwe.deserialize_json(data, bob_key, sender_key=alice_key)


def test_generic_serialize_deserialize_for_compact_serialization():
    jwe = JsonWebEncryption()

    alice_key = OKPKey.generate_key("X25519", is_private=True)
    bob_key = OKPKey.generate_key("X25519", is_private=True)

    header_obj = {"alg": "ECDH-1PU+A128KW", "enc": "A128CBC-HS256"}

    data = jwe.serialize(header_obj, b"hello", bob_key, sender_key=alice_key)
    assert isinstance(data, bytes)

    rv = jwe.deserialize(data, bob_key, sender_key=alice_key)
    assert rv["payload"] == b"hello"


def test_generic_serialize_deserialize_for_json_serialization():
    jwe = JsonWebEncryption()

    alice_key = OKPKey.generate_key("X25519", is_private=True)
    bob_key = OKPKey.generate_key("X25519", is_private=True)

    protected = {"alg": "ECDH-1PU+A128KW", "enc": "A128CBC-HS256"}
    header_obj = {"protected": protected}

    data = jwe.serialize(header_obj, b"hello", bob_key, sender_key=alice_key)
    assert isinstance(data, dict)

    rv = jwe.deserialize(data, bob_key, sender_key=alice_key)
    assert rv["payload"] == b"hello"


def test_generic_deserialize_for_json_serialization_string():
    jwe = JsonWebEncryption()

    alice_key = OKPKey.generate_key("X25519", is_private=True)
    bob_key = OKPKey.generate_key("X25519", is_private=True)

    protected = {"alg": "ECDH-1PU+A128KW", "enc": "A128CBC-HS256"}
    header_obj = {"protected": protected}

    data = jwe.serialize(header_obj, b"hello", bob_key, sender_key=alice_key)
    assert isinstance(data, dict)

    data_as_string = json.dumps(data)

    rv = jwe.deserialize(data_as_string, bob_key, sender_key=alice_key)
    assert rv["payload"] == b"hello"



"""authlib.oauth1.rfc5849.errors.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

RFC5849 has no definition on errors. This module is designed by
Authlib based on OAuth 1.0a `Section 10`_ with some changes.

.. _`Section 10`: https://oauth.net/core/1.0a/#rfc.section.10
"""

from authlib.common.errors import AuthlibHTTPError
from authlib.common.security import is_secure_transport


class OAuth1Error(AuthlibHTTPError):
    def __init__(self, description=None, uri=None, status_code=None):
        super().__init__(None, description, uri, status_code)

    def get_headers(self):
        """Get a list of headers."""
        return [
            ("Content-Type", "application/x-www-form-urlencoded"),
            ("Cache-Control", "no-store"),
            ("Pragma", "no-cache"),
        ]


class InsecureTransportError(OAuth1Error):
    error = "insecure_transport"
    description = "OAuth 2 MUST utilize https."

    @classmethod
    def check(cls, uri):
        if not is_secure_transport(uri):
            raise cls()


class InvalidRequestError(OAuth1Error):
    error = "invalid_request"


class UnsupportedParameterError(OAuth1Error):
    error = "unsupported_parameter"


class UnsupportedSignatureMethodError(OAuth1Error):
    error = "unsupported_signature_method"


class MissingRequiredParameterError(OAuth1Error):
    error = "missing_required_parameter"

    def __init__(self, key):
        description = f'missing "{key}" in parameters'
        super().__init__(description=description)


class DuplicatedOAuthProtocolParameterError(OAuth1Error):
    error = "duplicated_oauth_protocol_parameter"


class InvalidClientError(OAuth1Error):
    error = "invalid_client"
    status_code = 401


class InvalidTokenError(OAuth1Error):
    error = "invalid_token"
    description = 'Invalid or expired "oauth_token" in parameters'
    status_code = 401


class InvalidSignatureError(OAuth1Error):
    error = "invalid_signature"
    status_code = 401


class InvalidNonceError(OAuth1Error):
    error = "invalid_nonce"
    status_code = 401


class AccessDeniedError(OAuth1Error):
    error = "access_denied"
    description = "The resource owner or authorization server denied the request"


class MethodNotAllowedError(OAuth1Error):
    error = "method_not_allowed"
    status_code = 405



import binascii
import os
import time

from authlib.common.security import generate_token
from authlib.consts import default_json_headers
from authlib.deprecate import deprecate
from authlib.jose import JoseError
from authlib.jose import JsonWebToken

from ..rfc6749 import AccessDeniedError
from ..rfc6749 import InvalidRequestError
from .claims import ClientMetadataClaims
from .errors import InvalidClientMetadataError
from .errors import InvalidSoftwareStatementError
from .errors import UnapprovedSoftwareStatementError


class ClientRegistrationEndpoint:
    """The client registration endpoint is an OAuth 2.0 endpoint designed to
    allow a client to be registered with the authorization server.
    """

    ENDPOINT_NAME = "client_registration"

    #: Rewrite this value with a list to support ``software_statement``
    #: e.g. ``software_statement_alg_values_supported = ['RS256']``
    software_statement_alg_values_supported = None

    def __init__(self, server=None, claims_classes=None):
        self.server = server
        self.claims_classes = claims_classes or [ClientMetadataClaims]

    def __call__(self, request):
        return self.create_registration_response(request)

    def create_registration_response(self, request):
        token = self.authenticate_token(request)
        if not token:
            raise AccessDeniedError()

        request.credential = token

        client_metadata = self.extract_client_metadata(request)
        client_info = self.generate_client_info(request)
        body = {}
        body.update(client_metadata)
        body.update(client_info)
        client = self.save_client(client_info, client_metadata, request)
        registration_info = self.generate_client_registration_info(client, request)
        if registration_info:
            body.update(registration_info)
        return 201, body, default_json_headers

    def extract_client_metadata(self, request):
        if not request.payload.data:
            raise InvalidRequestError()

        json_data = request.payload.data.copy()
        software_statement = json_data.pop("software_statement", None)
        if software_statement and self.software_statement_alg_values_supported:
            data = self.extract_software_statement(software_statement, request)
            json_data.update(data)

        client_metadata = {}
        server_metadata = self.get_server_metadata()
        for claims_class in self.claims_classes:
            options = (
                claims_class.get_claims_options(server_metadata)
                if hasattr(claims_class, "get_claims_options") and server_metadata
                else {}
            )
            claims = claims_class(json_data, {}, options, server_metadata)
            try:
                claims.validate()
            except JoseError as error:
                raise InvalidClientMetadataError(error.description) from error

            client_metadata.update(**claims.get_registered_claims())
        return client_metadata

    def extract_software_statement(self, software_statement, request):
        key = self.resolve_public_key(request)
        if not key:
            raise UnapprovedSoftwareStatementError()

        try:
            jwt = JsonWebToken(self.software_statement_alg_values_supported)
            claims = jwt.decode(software_statement, key)
            # there is no need to validate claims
            return claims
        except JoseError as exc:
            raise InvalidSoftwareStatementError() from exc

    def generate_client_info(self, request):
        # https://tools.ietf.org/html/rfc7591#section-3.2.1
        try:
            client_id = self.generate_client_id(request)
        except TypeError:  # pragma: no cover
            client_id = self.generate_client_id()
            deprecate(
                "generate_client_id takes a 'request' parameter. "
                "It will become mandatory in coming releases",
                version="1.8",
            )

        try:
            client_secret = self.generate_client_secret(request)
        except TypeError:  # pragma: no cover
            client_secret = self.generate_client_secret()
            deprecate(
                "generate_client_secret takes a 'request' parameter. "
                "It will become mandatory in coming releases",
                version="1.8",
            )

        client_id_issued_at = int(time.time())
        client_secret_expires_at = 0
        return dict(
            client_id=client_id,
            client_secret=client_secret,
            client_id_issued_at=client_id_issued_at,
            client_secret_expires_at=client_secret_expires_at,
        )

    def generate_client_registration_info(self, client, request):
        """Generate ```registration_client_uri`` and ``registration_access_token``
        for RFC7592. This method returns ``None`` by default. Developers MAY rewrite
        this method to return registration information.
        """
        return None

    def create_endpoint_request(self, request):
        return self.server.create_json_request(request)

    def generate_client_id(self, request):
        """Generate ``client_id`` value. Developers MAY rewrite this method
        to use their own way to generate ``client_id``.
        """
        return generate_token(42)

    def generate_client_secret(self, request):
        """Generate ``client_secret`` value. Developers MAY rewrite this method
        to use their own way to generate ``client_secret``.
        """
        return binascii.hexlify(os.urandom(24)).decode("ascii")

    def get_server_metadata(self):
        """Return server metadata which includes supported grant types,
        response types and etc.
        """
        raise NotImplementedError()

    def authenticate_token(self, request):
        """Authenticate current credential who is requesting to register a client.
        Developers MUST implement this method in subclass::

            def authenticate_token(self, request):
                auth = request.headers.get("Authorization")
                return get_token_by_auth(auth)

        :return: token instance
        """
        raise NotImplementedError()

    def resolve_public_key(self, request):
        """Resolve a public key for decoding ``software_statement``. If
        ``enable_software_statement=True``, developers MUST implement this
        method in subclass::

            def resolve_public_key(self, request):
                return get_public_key_from_user(request.credential)

        :return: JWK or Key string
        """
        raise NotImplementedError()

    def save_client(self, client_info, client_metadata, request):
        """Save client into database. Developers MUST implement this method
        in subclass::

            def save_client(self, client_info, client_metadata, request):
                client = OAuthClient(
                    client_id=client_info['client_id'],
                    client_secret=client_info['client_secret'],
                    ...
                )
                client.save()
                return client
        """
        raise NotImplementedError()



import time

import pytest

from authlib.oauth1.rfc5849 import signature
from tests.util import decode_response
from tests.util import read_file_path

from .oauth1_server import Client
from .oauth1_server import User
from .oauth1_server import create_authorization_server


@pytest.fixture(autouse=True)
def user(db):
    user = User(username="foo")
    db.session.add(user)
    db.session.commit()
    yield user
    db.session.delete(user)


@pytest.fixture(autouse=True)
def client(db, user):
    client = Client(
        user_id=user.id,
        client_id="client",
        client_secret="secret",
        default_redirect_uri="https://client.test",
    )
    db.session.add(client)
    db.session.commit()
    yield db
    db.session.delete(client)


@pytest.mark.parametrize("use_cache", [True, False])
def test_temporary_credential_parameters_errors(app, test_client, use_cache):
    create_authorization_server(app, use_cache)
    url = "/oauth/initiate"

    rv = test_client.get(url)
    data = decode_response(rv.data)
    assert data["error"] == "method_not_allowed"

    # case 1
    rv = test_client.post(url)
    data = decode_response(rv.data)
    assert data["error"] == "missing_required_parameter"
    assert "oauth_consumer_key" in data["error_description"]

    # case 2
    rv = test_client.post(url, data={"oauth_consumer_key": "client"})
    data = decode_response(rv.data)
    assert data["error"] == "missing_required_parameter"
    assert "oauth_callback" in data["error_description"]

    # case 3
    rv = test_client.post(
        url, data={"oauth_consumer_key": "client", "oauth_callback": "invalid_url"}
    )
    data = decode_response(rv.data)
    assert data["error"] == "invalid_request"
    assert "oauth_callback" in data["error_description"]

    # case 4
    rv = test_client.post(
        url, data={"oauth_consumer_key": "invalid-client", "oauth_callback": "oob"}
    )
    data = decode_response(rv.data)
    assert data["error"] == "invalid_client"


@pytest.mark.parametrize("use_cache", [True, False])
def test_validate_timestamp_and_nonce(app, test_client, use_cache):
    create_authorization_server(app, use_cache)
    url = "/oauth/initiate"

    # case 5
    rv = test_client.post(
        url, data={"oauth_consumer_key": "client", "oauth_callback": "oob"}
    )
    data = decode_response(rv.data)
    assert data["error"] == "missing_required_parameter"
    assert "oauth_timestamp" in data["error_description"]

    # case 6
    rv = test_client.post(
        url,
        data={
            "oauth_consumer_key": "client",
            "oauth_callback": "oob",
            "oauth_timestamp": str(int(time.time())),
        },
    )
    data = decode_response(rv.data)
    assert data["error"] == "missing_required_parameter"
    assert "oauth_nonce" in data["error_description"]

    # case 7
    rv = test_client.post(
        url,
        data={
            "oauth_consumer_key": "client",
            "oauth_callback": "oob",
            "oauth_timestamp": "123",
        },
    )
    data = decode_response(rv.data)
    assert data["error"] == "invalid_request"
    assert "oauth_timestamp" in data["error_description"]

    # case 8
    rv = test_client.post(
        url,
        data={
            "oauth_consumer_key": "client",
            "oauth_callback": "oob",
            "oauth_timestamp": "sss",
        },
    )
    data = decode_response(rv.data)
    assert data["error"] == "invalid_request"
    assert "oauth_timestamp" in data["error_description"]

    # case 9
    rv = test_client.post(
        url,
        data={
            "oauth_consumer_key": "client",
            "oauth_callback": "oob",
            "oauth_timestamp": "-1",
            "oauth_signature_method": "PLAINTEXT",
        },
    )
    assert data["error"] == "invalid_request"
    assert "oauth_timestamp" in data["error_description"]


@pytest.mark.parametrize("use_cache", [True, False])
def test_temporary_credential_signatures_errors(app, test_client, use_cache):
    create_authorization_server(app, use_cache)
    url = "/oauth/initiate"

    rv = test_client.post(
        url,
        data={
            "oauth_consumer_key": "client",
            "oauth_callback": "oob",
            "oauth_signature_method": "PLAINTEXT",
        },
    )
    data = decode_response(rv.data)
    assert data["error"] == "missing_required_parameter"
    assert "oauth_signature" in data["error_description"]

    rv = test_client.post(
        url,
        data={
            "oauth_consumer_key": "client",
            "oauth_callback": "oob",
            "oauth_timestamp": str(int(time.time())),
            "oauth_nonce": "a",
        },
    )
    data = decode_response(rv.data)
    assert data["error"] == "missing_required_parameter"
    assert "oauth_signature_method" in data["error_description"]

    rv = test_client.post(
        url,
        data={
            "oauth_consumer_key": "client",
            "oauth_signature_method": "INVALID",
            "oauth_callback": "oob",
            "oauth_timestamp": str(int(time.time())),
            "oauth_nonce": "b",
            "oauth_signature": "c",
        },
    )
    data = decode_response(rv.data)
    assert data["error"] == "unsupported_signature_method"


@pytest.mark.parametrize("use_cache", [True, False])
def test_plaintext_signature(app, test_client, use_cache):
    create_authorization_server(app, use_cache)
    url = "/oauth/initiate"

    # case 1: use payload
    rv = test_client.post(
        url,
        data={
            "oauth_consumer_key": "client",
            "oauth_callback": "oob",
            "oauth_signature_method": "PLAINTEXT",
            "oauth_signature": "secret&",
        },
    )
    data = decode_response(rv.data)
    assert "oauth_token" in data

    # case 2: use header
    auth_header = (
        'OAuth oauth_consumer_key="client",'
        'oauth_signature_method="PLAINTEXT",'
        'oauth_callback="oob",'
        'oauth_signature="secret&"'
    )
    headers = {"Authorization": auth_header}
    rv = test_client.post(url, headers=headers)
    data = decode_response(rv.data)
    assert "oauth_token" in data

    # case 3: invalid signature
    rv = test_client.post(
        url,
        data={
            "oauth_consumer_key": "client",
            "oauth_callback": "oob",
            "oauth_signature_method": "PLAINTEXT",
            "oauth_signature": "invalid-signature",
        },
    )
    data = decode_response(rv.data)
    assert data["error"] == "invalid_signature"


@pytest.mark.parametrize("use_cache", [True, False])
def test_hmac_sha1_signature(app, test_client, use_cache):
    create_authorization_server(app, use_cache)
    url = "/oauth/initiate"

    params = [
        ("oauth_consumer_key", "client"),
        ("oauth_callback", "oob"),
        ("oauth_signature_method", "HMAC-SHA1"),
        ("oauth_timestamp", str(int(time.time()))),
        ("oauth_nonce", "hmac-sha1-nonce"),
    ]
    base_string = signature.construct_base_string(
        "POST", "http://localhost/oauth/initiate", params
    )
    sig = signature.hmac_sha1_signature(base_string, "secret", None)
    params.append(("oauth_signature", sig))
    auth_param = ",".join([f'{k}="{v}"' for k, v in params])
    auth_header = "OAuth " + auth_param
    headers = {"Authorization": auth_header}

    # case 1: success
    rv = test_client.post(url, headers=headers)
    data = decode_response(rv.data)
    assert "oauth_token" in data

    # case 2: exists nonce
    rv = test_client.post(url, headers=headers)
    data = decode_response(rv.data)
    assert data["error"] == "invalid_nonce"


@pytest.mark.parametrize("use_cache", [True, False])
def test_rsa_sha1_signature(app, test_client, use_cache):
    create_authorization_server(app, use_cache)
    url = "/oauth/initiate"

    params = [
        ("oauth_consumer_key", "client"),
        ("oauth_callback", "oob"),
        ("oauth_signature_method", "RSA-SHA1"),
        ("oauth_timestamp", str(int(time.time()))),
        ("oauth_nonce", "rsa-sha1-nonce"),
    ]
    base_string = signature.construct_base_string(
        "POST", "http://localhost/oauth/initiate", params
    )
    sig = signature.rsa_sha1_signature(base_string, read_file_path("rsa_private.pem"))
    params.append(("oauth_signature", sig))
    auth_param = ",".join([f'{k}="{v}"' for k, v in params])
    auth_header = "OAuth " + auth_param
    headers = {"Authorization": auth_header}
    rv = test_client.post(url, headers=headers)
    data = decode_response(rv.data)
    assert "oauth_token" in data

    # case: invalid signature
    auth_param = auth_param.replace("rsa-sha1-nonce", "alt-sha1-nonce")
    auth_header = "OAuth " + auth_param
    headers = {"Authorization": auth_header}
    rv = test_client.post(url, headers=headers)
    data = decode_response(rv.data)
    assert data["error"] == "invalid_signature"


@pytest.mark.parametrize("use_cache", [True, False])
def test_invalid_signature(app, test_client, use_cache):
    app.config.update({"OAUTH1_SUPPORTED_SIGNATURE_METHODS": ["INVALID"]})
    create_authorization_server(app, use_cache)
    url = "/oauth/initiate"
    rv = test_client.post(
        url,
        data={
            "oauth_consumer_key": "client",
            "oauth_callback": "oob",
            "oauth_signature_method": "PLAINTEXT",
            "oauth_signature": "secret&",
        },
    )
    data = decode_response(rv.data)
    assert data["error"] == "unsupported_signature_method"

    rv = test_client.post(
        url,
        data={
            "oauth_consumer_key": "client",
            "oauth_callback": "oob",
            "oauth_signature_method": "INVALID",
            "oauth_timestamp": str(int(time.time())),
            "oauth_nonce": "invalid-nonce",
            "oauth_signature": "secret&",
        },
    )
    data = decode_response(rv.data)
    assert data["error"] == "unsupported_signature_method"


@pytest.mark.parametrize("use_cache", [True, False])
def test_register_signature_method(app, test_client, use_cache):
    server = create_authorization_server(app, use_cache)

    def foo():
        pass

    server.register_signature_method("foo", foo)
    assert server.SIGNATURE_METHODS["foo"] == foo



import logging
import warnings

from authlib.oauth2.rfc6749 import AccessDeniedError
from authlib.oauth2.rfc6749 import ImplicitGrant
from authlib.oauth2.rfc6749 import InvalidScopeError
from authlib.oauth2.rfc6749 import OAuth2Error
from authlib.oauth2.rfc6749.errors import InvalidRequestError
from authlib.oauth2.rfc6749.hooks import hooked

from .util import create_response_mode_response
from .util import generate_id_token
from .util import is_openid_scope
from .util import validate_nonce
from .util import validate_request_prompt

log = logging.getLogger(__name__)


class OpenIDImplicitGrant(ImplicitGrant):
    RESPONSE_TYPES = {"id_token token", "id_token"}
    DEFAULT_RESPONSE_MODE = "fragment"

    def exists_nonce(self, nonce, request):
        """Check if the given nonce is existing in your database. Developers
        should implement this method in subclass, e.g.::

            def exists_nonce(self, nonce, request):
                exists = AuthorizationCode.query.filter_by(
                    client_id=request.payload.client_id, nonce=nonce
                ).first()
                return bool(exists)

        :param nonce: A string of "nonce" parameter in request
        :param request: OAuth2Request instance
        :return: Boolean
        """
        raise NotImplementedError()

    def get_jwt_config(self, client):
        """Get the JWT configuration for OpenIDImplicitGrant. The JWT
        configuration will be used to generate ``id_token``. Developers
        MUST implement this method in subclass, e.g.::

            def get_jwt_config(self, client):
                return {
                    "key": read_private_key_file(key_path),
                    "alg": client.id_token_signed_response_alg or "RS256",
                    "iss": "issuer-identity",
                    "exp": 3600,
                }

        :param client: OAuth2 client instance
        :return: dict
        """
        raise NotImplementedError()

    def generate_user_info(self, user, scope):
        """Provide user information for the given scope. Developers
        MUST implement this method in subclass, e.g.::

            from authlib.oidc.core import UserInfo


            def generate_user_info(self, user, scope):
                user_info = UserInfo(sub=user.id, name=user.name)
                if "email" in scope:
                    user_info["email"] = user.email
                return user_info

        :param user: user instance
        :param scope: scope of the token
        :return: ``authlib.oidc.core.UserInfo`` instance
        """
        raise NotImplementedError()

    def get_audiences(self, request):
        """Parse `aud` value for id_token, default value is client id. Developers
        MAY rewrite this method to provide a customized audience value.
        """
        client = request.client
        return [client.get_client_id()]

    def validate_authorization_request(self):
        if not is_openid_scope(self.request.payload.scope):
            raise InvalidScopeError(
                "Missing 'openid' scope",
                redirect_uri=self.request.payload.redirect_uri,
                redirect_fragment=True,
            )
        redirect_uri = super().validate_authorization_request()
        try:
            validate_nonce(self.request, self.exists_nonce, required=True)
        except OAuth2Error as error:
            error.redirect_uri = redirect_uri
            error.redirect_fragment = True
            raise error
        return redirect_uri

    @hooked
    def validate_consent_request(self):
        redirect_uri = self.validate_authorization_request()
        validate_request_prompt(self, redirect_uri, redirect_fragment=True)
        return redirect_uri

    def create_authorization_response(self, redirect_uri, grant_user):
        state = self.request.payload.state
        if grant_user:
            params = self.create_granted_params(grant_user)
            if state:
                params.append(("state", state))
        else:
            error = AccessDeniedError()
            params = error.get_body()

        # http://openid.net/specs/oauth-v2-multiple-response-types-1_0.html#ResponseModes
        response_mode = self.request.payload.data.get(
            "response_mode", self.DEFAULT_RESPONSE_MODE
        )
        return create_response_mode_response(
            redirect_uri=redirect_uri,
            params=params,
            response_mode=response_mode,
        )

    def create_granted_params(self, grant_user):
        self.request.user = grant_user
        client = self.request.client
        token = self.generate_token(
            user=grant_user,
            scope=self.request.payload.scope,
            include_refresh_token=False,
        )
        if self.request.payload.response_type == "id_token":
            token = {
                "expires_in": token["expires_in"],
                "scope": token["scope"],
            }
            token = self.process_implicit_token(token)
        else:
            log.debug("Grant token %r to %r", token, client)
            self.server.save_token(token, self.request)
            token = self.process_implicit_token(token)
        params = [(k, token[k]) for k in token]
        return params

    def process_implicit_token(self, token, code=None):
        try:
            config = self.get_jwt_config(self.request.client)
        except TypeError:
            warnings.warn(
                "get_jwt_config(self) is deprecated and will be removed in version 1.8. "
                "Use get_jwt_config(self, client) instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            config = self.get_jwt_config()

        config["aud"] = self.get_audiences(self.request)
        config["nonce"] = self.request.payload.data.get("nonce")
        if code is not None:
            config["code"] = code

        # Per OpenID Connect Registration 1.0 Section 2:
        # Use client's id_token_signed_response_alg if specified
        if not config.get("alg") and (
            client_alg := self.request.client.id_token_signed_response_alg
        ):
            if client_alg == "none":
                # According to oidc-registration §2 the 'none' alg is not valid in
                # implicit flows:
                #    The value none MUST NOT be used as the ID Token alg value unless
                #    the Client uses only Response Types that return no ID Token from
                #    the Authorization Endpoint (such as when only using the
                #    Authorization Code Flow).
                raise InvalidRequestError(
                    "id_token must be signed in implicit flows",
                    redirect_uri=self.request.payload.redirect_uri,
                    redirect_fragment=True,
                )

            config["alg"] = client_alg

        user_info = self.generate_user_info(self.request.user, token["scope"])
        id_token = generate_id_token(token, user_info, **config)
        token["id_token"] = id_token
        return token



import authlib

project = "Authlib"
copyright = "&copy; 2017, Hsiaoming Ltd"
author = "Hsiaoming Yang"
version = authlib.__version__
release = version

templates_path = ["_templates"]
html_static_path = ["_static"]
html_css_files = [
    "custom.css",
]
html_theme = "shibuya"

html_copy_source = False
html_show_sourcelink = False

language = "en"

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.extlinks",
    "sphinx_copybutton",
    "sphinx_design",
]

extlinks = {
    "issue": ("https://github.com/authlib/authlib/issues/%s", "issue #%s"),
    "PR": ("https://github.com/authlib/authlib/pull/%s", "pull request #%s"),
}

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
}
html_favicon = "_static/icon.svg"
html_theme_options = {
    "accent_color": "blue",
    "og_image_url": "https://authlib.org/logo.png",
    "light_logo": "_static/light-logo.svg",
    "dark_logo": "_static/dark-logo.svg",
    "twitter_site": "authlib",
    "twitter_creator": "lepture",
    "twitter_url": "https://twitter.com/authlib",
    "github_url": "https://github.com/authlib/authlib",
    "discord_url": "https://discord.gg/HvBVAeNAaV",
    "nav_links": [
        {
            "title": "Projects",
            "children": [
                {
                    "title": "Authlib",
                    "url": "https://authlib.org/",
                    "summary": "OAuth, JOSE, OpenID, etc.",
                },
                {
                    "title": "JOSE RFC",
                    "url": "https://jose.authlib.org/",
                    "summary": "JWS, JWE, JWK, and JWT.",
                },
                {
                    "title": "OTP Auth",
                    "url": "https://otp.authlib.org/",
                    "summary": "One time password, HOTP/TOTP.",
                },
            ],
        },
        {"title": "Sponsor me", "url": "https://github.com/sponsors/lepture"},
    ],
}

html_context = {}



import time
from copy import deepcopy
from unittest import mock

import pytest

from authlib.common.security import generate_token
from authlib.common.urls import add_params_to_uri
from authlib.common.urls import url_encode
from authlib.integrations.requests_client import OAuth2Session
from authlib.integrations.requests_client import OAuthError
from authlib.oauth2.rfc6749 import MismatchingStateException
from authlib.oauth2.rfc7523 import ClientSecretJWT
from authlib.oauth2.rfc7523 import PrivateKeyJWT

from ..util import read_key_file


def mock_json_response(payload):
    def fake_send(r, **kwargs):
        resp = mock.MagicMock()
        resp.status_code = 200
        resp.json = lambda: payload
        return resp

    return fake_send


def mock_assertion_response(token, session):
    def fake_send(r, **kwargs):
        assert "client_assertion=" in r.body
        assert "client_assertion_type=" in r.body
        resp = mock.MagicMock()
        resp.status_code = 200
        resp.json = lambda: token
        return resp

    session.send = fake_send


@pytest.fixture
def token():
    return {
        "token_type": "Bearer",
        "access_token": "a",
        "refresh_token": "b",
        "expires_in": "3600",
        "expires_at": int(time.time()) + 3600,
    }


def test_invalid_token_type(token):
    token = {
        "token_type": "invalid",
        "access_token": "a",
        "refresh_token": "b",
        "expires_in": "3600",
        "expires_at": int(time.time()) + 3600,
    }
    with OAuth2Session("foo", token=token) as sess:
        with pytest.raises(OAuthError):
            sess.get("https://provider.test")


def test_add_token_to_header(token):
    expected_header = "Bearer " + token["access_token"]

    def verifier(r, **kwargs):
        auth_header = r.headers.get("Authorization", None)
        assert auth_header == expected_header
        resp = mock.MagicMock()
        return resp

    sess = OAuth2Session(client_id="foo", token=token)
    sess.send = verifier
    sess.get("https://provider.test")


def test_add_token_to_body(token):
    def verifier(r, **kwargs):
        assert token["access_token"] in r.body
        resp = mock.MagicMock()
        return resp

    sess = OAuth2Session(client_id="foo", token=token, token_placement="body")
    sess.send = verifier
    sess.post("https://provider.test")


def test_add_token_to_uri(token):
    def verifier(r, **kwargs):
        assert token["access_token"] in r.url
        resp = mock.MagicMock()
        return resp

    sess = OAuth2Session(client_id="foo", token=token, token_placement="uri")
    sess.send = verifier
    sess.get("https://provider.test")


def test_create_authorization_url():
    url = "https://provider.test/authorize?foo=bar"

    sess = OAuth2Session(client_id="foo")
    auth_url, state = sess.create_authorization_url(url)
    assert state in auth_url
    assert "foo" in auth_url
    assert "response_type=code" in auth_url

    sess = OAuth2Session(client_id="foo", prompt="none")
    auth_url, state = sess.create_authorization_url(
        url, state="foo", redirect_uri="https://provider.test", scope="profile"
    )
    assert state == "foo"
    assert "provider.test" in auth_url
    assert "profile" in auth_url
    assert "prompt=none" in auth_url


def test_code_challenge():
    sess = OAuth2Session(client_id="foo", code_challenge_method="S256")

    url = "https://provider.test/authorize"
    auth_url, _ = sess.create_authorization_url(url, code_verifier=generate_token(48))
    assert "code_challenge" in auth_url
    assert "code_challenge_method=S256" in auth_url


def test_token_from_fragment(token):
    sess = OAuth2Session("foo")
    response_url = "https://provider.test/callback#" + url_encode(token.items())
    assert sess.token_from_fragment(response_url) == token
    token = sess.fetch_token(authorization_response=response_url)
    assert token == token


def test_fetch_token_post(token):
    url = "https://provider.test/token"

    def fake_send(r, **kwargs):
        assert "code=v" in r.body
        assert "client_id=" in r.body
        assert "grant_type=authorization_code" in r.body
        resp = mock.MagicMock()
        resp.status_code = 200
        resp.json = lambda: token
        return resp

    sess = OAuth2Session(client_id="foo")
    sess.send = fake_send
    assert (
        sess.fetch_token(url, authorization_response="https://provider.test/?code=v")
        == token
    )

    sess = OAuth2Session(
        client_id="foo",
        token_endpoint_auth_method="none",
    )
    sess.send = fake_send
    token = sess.fetch_token(url, code="v")
    assert token == token

    error = {"error": "invalid_request"}
    sess = OAuth2Session(client_id="foo", token=token)
    sess.send = mock_json_response(error)
    with pytest.raises(OAuthError):
        sess.fetch_access_token(url)


def test_fetch_token_get(token):
    url = "https://provider.test/token"

    def fake_send(r, **kwargs):
        assert "code=v" in r.url
        assert "grant_type=authorization_code" in r.url
        resp = mock.MagicMock()
        resp.status_code = 200
        resp.json = lambda: token
        return resp

    sess = OAuth2Session(client_id="foo")
    sess.send = fake_send
    token = sess.fetch_token(
        url, authorization_response="https://provider.test/?code=v", method="GET"
    )
    assert token == token

    sess = OAuth2Session(
        client_id="foo",
        token_endpoint_auth_method="none",
    )
    sess.send = fake_send
    token = sess.fetch_token(url, code="v", method="GET")
    assert token == token

    token = sess.fetch_token(url + "?q=a", code="v", method="GET")
    assert token == token


def test_token_auth_method_client_secret_post(token):
    url = "https://provider.test/token"

    def fake_send(r, **kwargs):
        assert "code=v" in r.body
        assert "client_id=" in r.body
        assert "client_secret=bar" in r.body
        assert "grant_type=authorization_code" in r.body
        resp = mock.MagicMock()
        resp.status_code = 200
        resp.json = lambda: token
        return resp

    sess = OAuth2Session(
        client_id="foo",
        client_secret="bar",
        token_endpoint_auth_method="client_secret_post",
    )
    sess.send = fake_send
    token = sess.fetch_token(url, code="v")
    assert token == token


def test_access_token_response_hook(token):
    url = "https://provider.test/token"

    def access_token_response_hook(resp):
        assert resp.json() == token
        return resp

    sess = OAuth2Session(client_id="foo", token=token)
    sess.register_compliance_hook("access_token_response", access_token_response_hook)
    sess.send = mock_json_response(token)
    assert sess.fetch_token(url) == token


def test_password_grant_type(token):
    url = "https://provider.test/token"

    def fake_send(r, **kwargs):
        assert "username=v" in r.body
        assert "grant_type=password" in r.body
        assert "scope=profile" in r.body
        resp = mock.MagicMock()
        resp.status_code = 200
        resp.json = lambda: token
        return resp

    sess = OAuth2Session(client_id="foo", scope="profile")
    sess.send = fake_send
    token = sess.fetch_token(url, username="v", password="v")
    assert token == token


def test_client_credentials_type(token):
    url = "https://provider.test/token"

    def fake_send(r, **kwargs):
        assert "grant_type=client_credentials" in r.body
        assert "scope=profile" in r.body
        resp = mock.MagicMock()
        resp.status_code = 200
        resp.json = lambda: token
        return resp

    sess = OAuth2Session(
        client_id="foo",
        client_secret="v",
        scope="profile",
    )
    sess.send = fake_send
    token = sess.fetch_token(url)
    assert token == token


def test_cleans_previous_token_before_fetching_new_one(token):
    """Makes sure the previous token is cleaned before fetching a new one.
    The reason behind it is that, if the previous token is expired, this
    method shouldn't fail with a TokenExpiredError, since it's attempting
    to get a new one (which shouldn't be expired).
    """
    now = int(time.time())
    new_token = deepcopy(token)
    past = now - 7200
    token["expires_at"] = past
    new_token["expires_at"] = now + 3600
    url = "https://provider.test/token"

    with mock.patch("time.time", lambda: now):
        sess = OAuth2Session(client_id="foo", token=token)
        sess.send = mock_json_response(new_token)
        assert sess.fetch_token(url) == new_token


def test_mis_match_state(token):
    sess = OAuth2Session("foo")
    with pytest.raises(MismatchingStateException):
        sess.fetch_token(
            "https://provider.test/token",
            authorization_response="https://provider.test/no-state?code=abc",
            state="somestate",
        )


def test_token_status():
    token = dict(access_token="a", token_type="bearer", expires_at=100)
    sess = OAuth2Session("foo", token=token)

    assert sess.token.is_expired


def test_token_status2():
    token = dict(access_token="a", token_type="bearer", expires_in=10)
    sess = OAuth2Session("foo", token=token, leeway=15)

    assert sess.token.is_expired(sess.leeway)


def test_token_status3():
    token = dict(access_token="a", token_type="bearer", expires_in=10)
    sess = OAuth2Session("foo", token=token, leeway=5)

    assert not sess.token.is_expired(sess.leeway)


def test_expires_in_used_when_expires_at_unparseable():
    """Test that expires_in is used as fallback when expires_at is unparsable."""
    token = dict(
        access_token="a",
        token_type="bearer",
        expires_in=3600,  # 1 hour from now
        expires_at="2024-01-01T00:00:00Z",  # Unparsable - should fall back to expires_in
    )
    sess = OAuth2Session("foo", token=token)

    # The token should use expires_in since expires_at is unparsable
    # So it should be considered expired with leeway > 3600
    assert sess.token.is_expired(leeway=3700) is True
    # And not expired with leeway < 3600
    assert sess.token.is_expired(leeway=0) is False
    # expires_at should be calculated from expires_in
    assert isinstance(sess.token["expires_at"], int)


def test_unparseable_expires_at_returns_none():
    """Test that is_expired returns None when expires_at is unparsable and no expires_in."""
    token = dict(
        access_token="a",
        token_type="bearer",
        expires_at="2024-01-01T00:00:00Z",  # Unparsable date string
    )
    sess = OAuth2Session("foo", token=token)

    # Should return None since we can't determine expiration
    assert sess.token.is_expired() is None
    # The unparsable expires_at should be preserved in the token
    assert sess.token["expires_at"] == "2024-01-01T00:00:00Z"
    # No expires_in should be calculated
    assert "expires_in" not in sess.token


def test_token_expired():
    token = dict(access_token="a", token_type="bearer", expires_at=100)
    sess = OAuth2Session("foo", token=token)
    with pytest.raises(OAuthError):
        sess.get(
            "https://provider.test/token",
        )


def test_missing_token():
    sess = OAuth2Session("foo")
    with pytest.raises(OAuthError):
        sess.get(
            "https://provider.test/token",
        )


def test_register_compliance_hook(token):
    sess = OAuth2Session("foo")
    with pytest.raises(ValueError):
        sess.register_compliance_hook(
            "invalid_hook",
            lambda o: o,
        )

    def protected_request(url, headers, data):
        assert "Authorization" in headers
        return url, headers, data

    sess = OAuth2Session("foo", token=token)
    sess.register_compliance_hook(
        "protected_request",
        protected_request,
    )
    sess.send = mock_json_response({"name": "a"})
    sess.get("https://resource.test/user")


def test_auto_refresh_token(token):
    def _update_token(token_, refresh_token=None, access_token=None):
        assert refresh_token == "b"
        assert token == token_

    update_token = mock.Mock(side_effect=_update_token)
    old_token = dict(
        access_token="a", refresh_token="b", token_type="bearer", expires_at=100
    )
    sess = OAuth2Session(
        "foo",
        token=old_token,
        token_endpoint="https://provider.test/token",
        update_token=update_token,
    )
    sess.send = mock_json_response(token)
    sess.get("https://resource.test/user")
    assert update_token.called


def test_auto_refresh_token2(token):
    def _update_token(token_, refresh_token=None, access_token=None):
        assert access_token == "a"
        assert token == token_

    update_token = mock.Mock(side_effect=_update_token)
    old_token = dict(access_token="a", token_type="bearer", expires_at=100)

    sess = OAuth2Session(
        "foo",
        token=old_token,
        token_endpoint="https://provider.test/token",
        grant_type="client_credentials",
    )
    sess.send = mock_json_response(token)
    sess.get("https://resource.test/user")
    assert not update_token.called

    sess = OAuth2Session(
        "foo",
        token=old_token,
        token_endpoint="https://provider.test/token",
        grant_type="client_credentials",
        update_token=update_token,
    )
    sess.send = mock_json_response(token)
    sess.get("https://resource.test/user")
    assert update_token.called


def test_revoke_token():
    sess = OAuth2Session("a")
    answer = {"status": "ok"}
    sess.send = mock_json_response(answer)
    resp = sess.revoke_token("https://provider.test/token", "hi")
    assert resp.json() == answer
    resp = sess.revoke_token(
        "https://provider.test/token", "hi", token_type_hint="access_token"
    )
    assert resp.json() == answer

    def revoke_token_request(url, headers, data):
        assert url == "https://provider.test/token"
        return url, headers, data

    sess.register_compliance_hook(
        "revoke_token_request",
        revoke_token_request,
    )
    sess.revoke_token(
        "https://provider.test/token", "hi", body="", token_type_hint="access_token"
    )


def test_introspect_token():
    sess = OAuth2Session("a")
    answer = {
        "active": True,
        "client_id": "l238j323ds-23ij4",
        "username": "jdoe",
        "scope": "read write dolphin",
        "sub": "Z5O3upPC88QrAjx00dis",
        "aud": "https://resource.test/resource",
        "iss": "https://provider.test/",
        "exp": 1419356238,
        "iat": 1419350238,
    }
    sess.send = mock_json_response(answer)
    resp = sess.introspect_token("https://provider.test/token", "hi")
    assert resp.json() == answer


def test_client_secret_jwt(token):
    sess = OAuth2Session("id", "secret", token_endpoint_auth_method="client_secret_jwt")
    sess.register_client_auth_method(ClientSecretJWT())

    mock_assertion_response(token, sess)
    token = sess.fetch_token("https://provider.test/token")
    assert token == token


def test_client_secret_jwt2(token):
    sess = OAuth2Session(
        "id",
        "secret",
        token_endpoint_auth_method=ClientSecretJWT(),
    )
    mock_assertion_response(token, sess)
    token = sess.fetch_token("https://provider.test/token")
    assert token == token


def test_private_key_jwt(token):
    client_secret = read_key_file("rsa_private.pem")
    sess = OAuth2Session(
        "id", client_secret, token_endpoint_auth_method="private_key_jwt"
    )
    sess.register_client_auth_method(PrivateKeyJWT())
    mock_assertion_response(token, sess)
    token = sess.fetch_token("https://provider.test/token")
    assert token == token


def test_custom_client_auth_method(token):
    def auth_client(client, method, uri, headers, body):
        uri = add_params_to_uri(
            uri,
            [
                ("client_id", client.client_id),
                ("client_secret", client.client_secret),
            ],
        )
        uri = uri + "&" + body
        body = ""
        return uri, headers, body

    sess = OAuth2Session("id", "secret", token_endpoint_auth_method="client_secret_uri")
    sess.register_client_auth_method(("client_secret_uri", auth_client))

    def fake_send(r, **kwargs):
        assert "client_id=" in r.url
        assert "client_secret=" in r.url
        resp = mock.MagicMock()
        resp.status_code = 200
        resp.json = lambda: token
        return resp

    sess.send = fake_send
    token = sess.fetch_token("https://provider.test/token")
    assert token == token


def test_use_client_token_auth(token):
    import requests

    expected_header = "Bearer " + token["access_token"]

    def verifier(r, **kwargs):
        auth_header = r.headers.get("Authorization", None)
        assert auth_header == expected_header
        resp = mock.MagicMock()
        return resp

    client = OAuth2Session(client_id="foo", token=token)

    sess = requests.Session()
    sess.send = verifier
    sess.get("https://provider.test", auth=client.token_auth)


def test_use_default_request_timeout(token):
    expected_timeout = 15

    def verifier(r, **kwargs):
        timeout = kwargs.get("timeout")
        assert timeout == expected_timeout
        resp = mock.MagicMock()
        return resp

    client = OAuth2Session(
        client_id="foo",
        token=token,
        default_timeout=expected_timeout,
    )

    client.send = verifier
    client.request("GET", "https://provider.test", withhold_token=False)


def test_override_default_request_timeout(token):
    default_timeout = 15
    expected_timeout = 10

    def verifier(r, **kwargs):
        timeout = kwargs.get("timeout")
        assert timeout == expected_timeout
        resp = mock.MagicMock()
        return resp

    client = OAuth2Session(
        client_id="foo",
        token=token,
        default_timeout=default_timeout,
    )

    client.send = verifier
    client.request(
        "GET", "https://provider.test", withhold_token=False, timeout=expected_timeout
    )



from unittest import mock

import pytest
from cachelib import SimpleCache
from flask import Flask
from flask import session

from authlib.common.urls import url_decode
from authlib.common.urls import urlparse
from authlib.integrations.flask_client import FlaskOAuth2App
from authlib.integrations.flask_client import OAuth
from authlib.integrations.flask_client import OAuthError
from authlib.jose.rfc7517 import JsonWebKey
from authlib.oauth2.rfc6749.errors import MissingCodeException
from authlib.oidc.core.grants.util import generate_id_token

from ..util import get_bearer_token
from ..util import mock_send_value


def test_register_remote_app():
    app = Flask(__name__)
    oauth = OAuth(app)
    with pytest.raises(AttributeError):
        oauth.dev  # noqa:B018

    oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
    )
    assert oauth.dev.name == "dev"
    assert oauth.dev.client_id == "dev"


def test_register_conf_from_app():
    app = Flask(__name__)
    app.config.update(
        {
            "DEV_CLIENT_ID": "dev",
            "DEV_CLIENT_SECRET": "dev",
        }
    )
    oauth = OAuth(app)
    oauth.register("dev")
    assert oauth.dev.client_id == "dev"


def test_register_with_overwrite():
    app = Flask(__name__)
    app.config.update(
        {
            "DEV_CLIENT_ID": "dev-1",
            "DEV_CLIENT_SECRET": "dev",
            "DEV_ACCESS_TOKEN_PARAMS": {"foo": "foo-1"},
        }
    )
    oauth = OAuth(app)
    oauth.register(
        "dev", overwrite=True, client_id="dev", access_token_params={"foo": "foo"}
    )
    assert oauth.dev.client_id == "dev-1"
    assert oauth.dev.client_secret == "dev"
    assert oauth.dev.access_token_params["foo"] == "foo-1"


def test_init_app_later():
    app = Flask(__name__)
    app.config.update(
        {
            "DEV_CLIENT_ID": "dev",
            "DEV_CLIENT_SECRET": "dev",
        }
    )
    oauth = OAuth()
    remote = oauth.register("dev")
    with pytest.raises(RuntimeError):
        oauth.dev.client_id  # noqa:B018
    oauth.init_app(app)
    assert oauth.dev.client_id == "dev"
    assert remote.client_id == "dev"

    assert oauth.cache is None
    assert oauth.fetch_token is None
    assert oauth.update_token is None


def test_init_app_params():
    app = Flask(__name__)
    oauth = OAuth()
    oauth.init_app(app, SimpleCache())
    assert oauth.cache is not None
    assert oauth.update_token is None

    oauth.init_app(app, update_token=lambda o: o)
    assert oauth.update_token is not None


def test_create_client():
    app = Flask(__name__)
    oauth = OAuth(app)
    assert oauth.create_client("dev") is None
    oauth.register("dev", client_id="dev")
    assert oauth.create_client("dev") is not None


def test_register_oauth1_remote_app():
    app = Flask(__name__)
    oauth = OAuth(app)
    client_kwargs = dict(
        client_id="dev",
        client_secret="dev",
        request_token_url="https://provider.test/request-token",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        authorize_url="https://provider.test/authorize",
        fetch_request_token=lambda: None,
        save_request_token=lambda token: token,
    )
    oauth.register("dev", **client_kwargs)
    assert oauth.dev.name == "dev"
    assert oauth.dev.client_id == "dev"

    oauth = OAuth(app, cache=SimpleCache())
    oauth.register("dev", **client_kwargs)
    assert oauth.dev.name == "dev"
    assert oauth.dev.client_id == "dev"


def test_oauth1_authorize_cache():
    app = Flask(__name__)
    app.secret_key = "!"
    cache = SimpleCache()
    oauth = OAuth(app, cache=cache)

    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        request_token_url="https://provider.test/request-token",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        authorize_url="https://provider.test/authorize",
    )

    with app.test_request_context():
        with mock.patch("requests.sessions.Session.send") as send:
            send.return_value = mock_send_value("oauth_token=foo&oauth_verifier=baz")
            resp = client.authorize_redirect("https://client.test/callback")
            assert resp.status_code == 302
            url = resp.headers.get("Location")
            assert "oauth_token=foo" in url
            session_data = session["_state_dev_foo"]
            assert "exp" in session_data
            assert "data" not in session_data

    with app.test_request_context("/?oauth_token=foo"):
        with mock.patch("requests.sessions.Session.send") as send:
            session["_state_dev_foo"] = session_data
            send.return_value = mock_send_value("oauth_token=a&oauth_token_secret=b")
            token = client.authorize_access_token()
            assert token["oauth_token"] == "a"


def test_oauth1_authorize_session():
    app = Flask(__name__)
    app.secret_key = "!"
    oauth = OAuth(app)
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        request_token_url="https://provider.test/request-token",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        authorize_url="https://provider.test/authorize",
    )

    with app.test_request_context():
        with mock.patch("requests.sessions.Session.send") as send:
            send.return_value = mock_send_value("oauth_token=foo&oauth_verifier=baz")
            resp = client.authorize_redirect("https://client.test/callback")
            assert resp.status_code == 302
            url = resp.headers.get("Location")
            assert "oauth_token=foo" in url
            data = session["_state_dev_foo"]

    with app.test_request_context("/?oauth_token=foo"):
        session["_state_dev_foo"] = data
        with mock.patch("requests.sessions.Session.send") as send:
            send.return_value = mock_send_value("oauth_token=a&oauth_token_secret=b")
            token = client.authorize_access_token()
            assert token["oauth_token"] == "a"


def test_register_oauth2_remote_app():
    app = Flask(__name__)
    oauth = OAuth(app)
    oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        refresh_token_url="https://provider.test/token",
        authorize_url="https://provider.test/authorize",
        update_token=lambda name: "hi",
    )
    assert oauth.dev.name == "dev"
    session = oauth.dev._get_oauth_client()
    assert session.update_token is not None


def test_oauth2_authorize_cache():
    app = Flask(__name__)
    app.secret_key = "!"
    cache = SimpleCache()
    oauth = OAuth(app, cache=cache)
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        authorize_url="https://provider.test/authorize",
    )
    with app.test_request_context():
        resp = client.authorize_redirect("https://client.test/callback")
        assert resp.status_code == 302
        url = resp.headers.get("Location")
        assert "state=" in url
        state = dict(url_decode(urlparse.urlparse(url).query))["state"]
        assert state is not None
        session_data = session[f"_state_dev_{state}"]
        assert "exp" in session_data
        assert "data" not in session_data

    with app.test_request_context(path=f"/?code=a&state={state}"):
        # session is cleared in tests
        session[f"_state_dev_{state}"] = session_data

        with mock.patch("requests.sessions.Session.send") as send:
            send.return_value = mock_send_value(get_bearer_token())
            token = client.authorize_access_token()
            assert token["access_token"] == "a"

    with app.test_request_context():
        assert client.token is None


def test_oauth2_authorize_session():
    app = Flask(__name__)
    app.secret_key = "!"
    oauth = OAuth(app)
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        authorize_url="https://provider.test/authorize",
    )

    with app.test_request_context():
        resp = client.authorize_redirect("https://client.test/callback")
        assert resp.status_code == 302
        url = resp.headers.get("Location")
        assert "state=" in url
        state = dict(url_decode(urlparse.urlparse(url).query))["state"]
        assert state is not None
        session_data = session[f"_state_dev_{state}"]
        assert "exp" in session_data
        assert "data" in session_data

    with app.test_request_context(path=f"/?code=a&state={state}"):
        # session is cleared in tests
        session[f"_state_dev_{state}"] = session_data

        with mock.patch("requests.sessions.Session.send") as send:
            send.return_value = mock_send_value(get_bearer_token())
            token = client.authorize_access_token()
            assert token["access_token"] == "a"

    with app.test_request_context():
        assert client.token is None


def test_oauth2_authorize_access_denied():
    app = Flask(__name__)
    app.secret_key = "!"
    oauth = OAuth(app)
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        authorize_url="https://provider.test/authorize",
    )

    with app.test_request_context(
        path="/?error=access_denied&error_description=Not+Allowed"
    ):
        # session is cleared in tests
        with mock.patch("requests.sessions.Session.send"):
            with pytest.raises(OAuthError):
                client.authorize_access_token()


def test_oauth2_authorize_via_custom_client():
    class CustomRemoteApp(FlaskOAuth2App):
        OAUTH_APP_CONFIG = {"authorize_url": "https://provider.test/custom"}

    app = Flask(__name__)
    app.secret_key = "!"
    oauth = OAuth(app)
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        client_cls=CustomRemoteApp,
    )
    with app.test_request_context():
        resp = client.authorize_redirect("https://client.test/callback")
        assert resp.status_code == 302
        url = resp.headers.get("Location")
        assert url.startswith("https://provider.test/custom?")


def test_oauth2_authorize_with_metadata():
    app = Flask(__name__)
    app.secret_key = "!"
    oauth = OAuth(app)
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
    )
    with pytest.raises(RuntimeError):
        client.create_authorization_url(None)

    client = oauth.register(
        "dev2",
        client_id="dev",
        client_secret="dev",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        server_metadata_url="https://provider.test/.well-known/openid-configuration",
    )
    with mock.patch("requests.sessions.Session.send") as send:
        send.return_value = mock_send_value(
            {"authorization_endpoint": "https://provider.test/authorize"}
        )

        with app.test_request_context():
            resp = client.authorize_redirect("https://client.test/callback")
            assert resp.status_code == 302


def test_oauth2_authorize_code_challenge():
    app = Flask(__name__)
    app.secret_key = "!"
    oauth = OAuth(app)
    client = oauth.register(
        "dev",
        client_id="dev",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        authorize_url="https://provider.test/authorize",
        client_kwargs={"code_challenge_method": "S256"},
    )

    with app.test_request_context():
        resp = client.authorize_redirect("https://client.test/callback")
        assert resp.status_code == 302
        url = resp.headers.get("Location")
        assert "code_challenge=" in url
        assert "code_challenge_method=S256" in url

        state = dict(url_decode(urlparse.urlparse(url).query))["state"]
        assert state is not None
        data = session[f"_state_dev_{state}"]

        verifier = data["data"]["code_verifier"]
        assert verifier is not None

    def fake_send(sess, req, **kwargs):
        assert f"code_verifier={verifier}" in req.body
        return mock_send_value(get_bearer_token())

    path = f"/?code=a&state={state}"
    with app.test_request_context(path=path):
        # session is cleared in tests
        session[f"_state_dev_{state}"] = data

        with mock.patch("requests.sessions.Session.send", fake_send):
            token = client.authorize_access_token()
            assert token["access_token"] == "a"


def test_openid_authorize():
    app = Flask(__name__)
    app.secret_key = "!"
    oauth = OAuth(app)
    key = dict(JsonWebKey.import_key("secret", {"kid": "f", "kty": "oct"}))

    client = oauth.register(
        "dev",
        client_id="dev",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        authorize_url="https://provider.test/authorize",
        client_kwargs={"scope": "openid profile"},
        jwks={"keys": [key]},
    )

    with app.test_request_context():
        resp = client.authorize_redirect("https://client.test/callback")
        assert resp.status_code == 302

        url = resp.headers["Location"]
        query_data = dict(url_decode(urlparse.urlparse(url).query))

        state = query_data["state"]
        assert state is not None
        session_data = session[f"_state_dev_{state}"]
        nonce = session_data["data"]["nonce"]
        assert nonce is not None
        assert nonce == query_data["nonce"]

    token = get_bearer_token()
    token["id_token"] = generate_id_token(
        token,
        {"sub": "123"},
        key,
        alg="HS256",
        iss="https://provider.test",
        aud="dev",
        exp=3600,
        nonce=query_data["nonce"],
    )
    path = f"/?code=a&state={state}"
    with app.test_request_context(path=path):
        session[f"_state_dev_{state}"] = session_data
        with mock.patch("requests.sessions.Session.send") as send:
            send.return_value = mock_send_value(token)
            token = client.authorize_access_token()
            assert token["access_token"] == "a"
            assert "userinfo" in token


def test_oauth2_access_token_with_post():
    app = Flask(__name__)
    app.secret_key = "!"
    oauth = OAuth(app)
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        authorize_url="https://provider.test/authorize",
    )
    payload = {"code": "a", "state": "b"}
    with app.test_request_context(data=payload, method="POST"):
        session["_state_dev_b"] = {"data": payload}
        with mock.patch("requests.sessions.Session.send") as send:
            send.return_value = mock_send_value(get_bearer_token())
            token = client.authorize_access_token()
            assert token["access_token"] == "a"


def test_access_token_with_fetch_token():
    app = Flask(__name__)
    app.secret_key = "!"
    oauth = OAuth()

    token = get_bearer_token()
    oauth.init_app(app, fetch_token=lambda name: token)
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        authorize_url="https://provider.test/authorize",
    )

    def fake_send(sess, req, **kwargs):
        auth = req.headers["Authorization"]
        assert auth == "Bearer {}".format(token["access_token"])
        resp = mock.MagicMock()
        resp.text = "hi"
        resp.status_code = 200
        return resp

    with app.test_request_context():
        with mock.patch("requests.sessions.Session.send", fake_send):
            resp = client.get("/api/user")
            assert resp.text == "hi"

            # trigger ctx.authlib_client_oauth_token
            resp = client.get("/api/user")
            assert resp.text == "hi"


def test_request_with_refresh_token():
    app = Flask(__name__)
    app.secret_key = "!"
    oauth = OAuth()

    expired_token = {
        "token_type": "Bearer",
        "access_token": "expired-a",
        "refresh_token": "expired-b",
        "expires_in": "3600",
        "expires_at": 1566465749,
    }
    oauth.init_app(app, fetch_token=lambda name: expired_token)
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        refresh_token_url="https://provider.test/token",
        authorize_url="https://provider.test/authorize",
    )

    def fake_send(sess, req, **kwargs):
        if req.url == "https://provider.test/token":
            auth = req.headers["Authorization"]
            assert "Basic" in auth
            resp = mock.MagicMock()
            resp.json = get_bearer_token
            resp.status_code = 200
            return resp

        resp = mock.MagicMock()
        resp.text = "hi"
        resp.status_code = 200
        return resp

    with app.test_request_context():
        with mock.patch("requests.sessions.Session.send", fake_send):
            resp = client.get("/api/user", token=expired_token)
            assert resp.text == "hi"


def test_request_without_token():
    app = Flask(__name__)
    app.secret_key = "!"
    oauth = OAuth(app)
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        authorize_url="https://provider.test/authorize",
    )

    def fake_send(sess, req, **kwargs):
        auth = req.headers.get("Authorization")
        assert auth is None
        resp = mock.MagicMock()
        resp.text = "hi"
        resp.status_code = 200
        return resp

    with app.test_request_context():
        with mock.patch("requests.sessions.Session.send", fake_send):
            resp = client.get("/api/user", withhold_token=True)
            assert resp.text == "hi"
            with pytest.raises(OAuthError):
                client.get("https://resource.test/api/user")


def test_oauth2_authorize_missing_code():
    app = Flask(__name__)
    app.secret_key = "!"
    oauth = OAuth(app)
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        authorize_url="https://provider.test/authorize",
    )

    with app.test_request_context():
        resp = client.authorize_redirect("https://client.test/callback")
        state = dict(url_decode(urlparse.urlparse(resp.headers["Location"]).query))[
            "state"
        ]
        session_data = session[f"_state_dev_{state}"]

    # Test missing code parameter
    with app.test_request_context(path=f"/?state={state}"):
        session[f"_state_dev_{state}"] = session_data
        with pytest.raises(MissingCodeException) as exc_info:
            client.authorize_access_token()
        assert exc_info.value.error == "missing_code"



from authlib.oidc.discovery.models import _validate_boolean_value


class AuthorizationServerMetadata(dict):
    REGISTRY_KEYS = ["require_signed_request_object"]

    def validate_require_signed_request_object(self):
        """Indicates where authorization request needs to be protected as Request Object and provided through either request or request_uri parameter."""
        _validate_boolean_value(self, "require_signed_request_object")



from unittest import mock

import pytest
from django.test import override_settings

from authlib.common.urls import url_decode
from authlib.common.urls import urlparse
from authlib.integrations.django_client import OAuth
from authlib.integrations.django_client import OAuthError
from authlib.jose import JsonWebKey
from authlib.oidc.core.grants.util import generate_id_token

from ..util import get_bearer_token
from ..util import mock_send_value

dev_client = {"client_id": "dev-key", "client_secret": "dev-secret"}


def test_register_remote_app():
    oauth = OAuth()
    with pytest.raises(AttributeError):
        oauth.dev  # noqa:B018

    oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        request_token_url="https://provider.test/request-token",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        authorize_url="https://provider.test/authorize",
    )
    assert oauth.dev.name == "dev"
    assert oauth.dev.client_id == "dev"


def test_register_with_overwrite():
    oauth = OAuth()
    oauth.register(
        "dev_overwrite",
        overwrite=True,
        client_id="dev",
        client_secret="dev",
        request_token_url="https://provider.test/request-token",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        access_token_params={"foo": "foo"},
        authorize_url="https://provider.test/authorize",
    )
    assert oauth.dev_overwrite.client_id == "dev-client-id"
    assert oauth.dev_overwrite.access_token_params["foo"] == "foo-1"


@override_settings(AUTHLIB_OAUTH_CLIENTS={"dev": dev_client})
def test_register_from_settings():
    oauth = OAuth()
    oauth.register("dev")
    assert oauth.dev.client_id == "dev-key"
    assert oauth.dev.client_secret == "dev-secret"


def test_oauth1_authorize(factory):
    request = factory.get("/login")
    request.session = factory.session

    oauth = OAuth()
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        request_token_url="https://provider.test/request-token",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        authorize_url="https://provider.test/authorize",
    )

    with mock.patch("requests.sessions.Session.send") as send:
        send.return_value = mock_send_value("oauth_token=foo&oauth_verifier=baz")

        resp = client.authorize_redirect(request)
        assert resp.status_code == 302
        url = resp.get("Location")
        assert "oauth_token=foo" in url

    request2 = factory.get(url)
    request2.session = request.session
    with mock.patch("requests.sessions.Session.send") as send:
        send.return_value = mock_send_value("oauth_token=a&oauth_token_secret=b")
        token = client.authorize_access_token(request2)
        assert token["oauth_token"] == "a"


def test_oauth2_authorize(factory):
    request = factory.get("/login")
    request.session = factory.session

    oauth = OAuth()
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        authorize_url="https://provider.test/authorize",
    )
    rv = client.authorize_redirect(request, "https://client.test/callback")
    assert rv.status_code == 302
    url = rv.get("Location")
    assert "state=" in url
    state = dict(url_decode(urlparse.urlparse(url).query))["state"]

    with mock.patch("requests.sessions.Session.send") as send:
        send.return_value = mock_send_value(get_bearer_token())
        request2 = factory.get(f"/authorize?state={state}&code=foo")
        request2.session = request.session

        token = client.authorize_access_token(request2)
        assert token["access_token"] == "a"


def test_oauth2_authorize_access_denied(factory):
    oauth = OAuth()
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        authorize_url="https://provider.test/authorize",
    )

    with mock.patch("requests.sessions.Session.send"):
        request = factory.get("/?error=access_denied&error_description=Not+Allowed")
        request.session = factory.session
        with pytest.raises(OAuthError):
            client.authorize_access_token(request)


def test_oauth2_authorize_code_challenge(factory):
    request = factory.get("/login")
    request.session = factory.session

    oauth = OAuth()
    client = oauth.register(
        "dev",
        client_id="dev",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        authorize_url="https://provider.test/authorize",
        client_kwargs={"code_challenge_method": "S256"},
    )
    rv = client.authorize_redirect(request, "https://client.test/callback")
    assert rv.status_code == 302
    url = rv.get("Location")
    assert "state=" in url
    assert "code_challenge=" in url

    state = dict(url_decode(urlparse.urlparse(url).query))["state"]
    state_data = request.session[f"_state_dev_{state}"]["data"]
    verifier = state_data["code_verifier"]

    def fake_send(sess, req, **kwargs):
        assert f"code_verifier={verifier}" in req.body
        return mock_send_value(get_bearer_token())

    with mock.patch("requests.sessions.Session.send", fake_send):
        request2 = factory.get(f"/authorize?state={state}&code=foo")
        request2.session = request.session
        token = client.authorize_access_token(request2)
        assert token["access_token"] == "a"


def test_oauth2_authorize_code_verifier(factory):
    request = factory.get("/login")
    request.session = factory.session

    oauth = OAuth()
    client = oauth.register(
        "dev",
        client_id="dev",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        authorize_url="https://provider.test/authorize",
        client_kwargs={"code_challenge_method": "S256"},
    )
    state = "foo"
    code_verifier = "bar"
    rv = client.authorize_redirect(
        request,
        "https://client.test/callback",
        state=state,
        code_verifier=code_verifier,
    )
    assert rv.status_code == 302
    url = rv.get("Location")
    assert "state=" in url
    assert "code_challenge=" in url

    with mock.patch("requests.sessions.Session.send") as send:
        send.return_value = mock_send_value(get_bearer_token())

        request2 = factory.get(f"/authorize?state={state}&code=foo")
        request2.session = request.session

        token = client.authorize_access_token(request2)
        assert token["access_token"] == "a"


def test_openid_authorize(factory):
    request = factory.get("/login")
    request.session = factory.session
    secret_key = JsonWebKey.import_key("secret", {"kty": "oct", "kid": "f"})

    oauth = OAuth()
    client = oauth.register(
        "dev",
        client_id="dev",
        jwks={"keys": [secret_key.as_dict()]},
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        authorize_url="https://provider.test/authorize",
        client_kwargs={"scope": "openid profile"},
    )

    resp = client.authorize_redirect(request, "https://client.test/callback")
    assert resp.status_code == 302
    url = resp.get("Location")
    assert "nonce=" in url
    query_data = dict(url_decode(urlparse.urlparse(url).query))

    token = get_bearer_token()
    token["id_token"] = generate_id_token(
        token,
        {"sub": "123"},
        secret_key,
        alg="HS256",
        iss="https://provider.test",
        aud="dev",
        exp=3600,
        nonce=query_data["nonce"],
    )
    state = query_data["state"]
    with mock.patch("requests.sessions.Session.send") as send:
        send.return_value = mock_send_value(token)

        request2 = factory.get(f"/authorize?state={state}&code=foo")
        request2.session = request.session

        token = client.authorize_access_token(request2)
        assert token["access_token"] == "a"
        assert "userinfo" in token
        assert token["userinfo"]["sub"] == "123"


def test_oauth2_access_token_with_post(factory):
    oauth = OAuth()
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        authorize_url="https://provider.test/authorize",
    )
    payload = {"code": "a", "state": "b"}

    with mock.patch("requests.sessions.Session.send") as send:
        send.return_value = mock_send_value(get_bearer_token())
        request = factory.post("/token", data=payload)
        request.session = factory.session
        request.session["_state_dev_b"] = {"data": {}}
        token = client.authorize_access_token(request)
        assert token["access_token"] == "a"


def test_with_fetch_token_in_oauth(factory):
    def fetch_token(name, request):
        return {"access_token": name, "token_type": "bearer"}

    oauth = OAuth(fetch_token=fetch_token)
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        authorize_url="https://provider.test/authorize",
    )

    def fake_send(sess, req, **kwargs):
        assert sess.token["access_token"] == "dev"
        return mock_send_value(get_bearer_token())

    with mock.patch("requests.sessions.Session.send", fake_send):
        request = factory.get("/login")
        client.get("/user", request=request)


def test_with_fetch_token_in_register(factory):
    def fetch_token(request):
        return {"access_token": "dev", "token_type": "bearer"}

    oauth = OAuth()
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        authorize_url="https://provider.test/authorize",
        fetch_token=fetch_token,
    )

    def fake_send(sess, req, **kwargs):
        assert sess.token["access_token"] == "dev"
        return mock_send_value(get_bearer_token())

    with mock.patch("requests.sessions.Session.send", fake_send):
        request = factory.get("/login")
        client.get("/user", request=request)


def test_request_without_token():
    oauth = OAuth()
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        authorize_url="https://provider.test/authorize",
    )

    def fake_send(sess, req, **kwargs):
        auth = req.headers.get("Authorization")
        assert auth is None
        resp = mock.MagicMock()
        resp.text = "hi"
        resp.status_code = 200
        return resp

    with mock.patch("requests.sessions.Session.send", fake_send):
        resp = client.get("/api/user", withhold_token=True)
        assert resp.text == "hi"
        with pytest.raises(OAuthError):
            client.get("https://resource.test/api/user")
